#!/usr/bin/env python
# Copyright (c) 2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import io
import json
from unittest import mock

from urllib.parse import parse_qs
from swift.common import swob
from swift.common.middleware import symlink, copy, versioned_writes, \
    listing_formats
from swift.common.swob import Request
from swift.common.request_helpers import get_reserved_name
from swift.common.utils import MD5_OF_EMPTY_STRING
from swift.common.registry import get_swift_info
from test.unit.common.middleware.helpers import FakeSwift
from test.unit.common.middleware.test_versioned_writes import FakeCache


class TestSymlinkMiddlewareBase(unittest.TestCase):
    def setUp(self):
        self.app = FakeSwift()
        self.sym = symlink.filter_factory({
            'symloop_max': '2',
        })(self.app)
        self.sym.logger = self.app.logger

    def call_app(self, req, app=None, expect_exception=False):
        if app is None:
            app = self.app

        self.authorized = []

        def authorize(req):
            self.authorized.append(req)

        if 'swift.authorize' not in req.environ:
            req.environ['swift.authorize'] = authorize

        status = [None]
        headers = [None]

        def start_response(s, h, ei=None):
            status[0] = s
            headers[0] = h

        body_iter = app(req.environ, start_response)
        body = b''
        caught_exc = None
        try:
            for chunk in body_iter:
                body += chunk
        except Exception as exc:
            if expect_exception:
                caught_exc = exc
            else:
                raise

        if expect_exception:
            return status[0], headers[0], body, caught_exc
        else:
            return status[0], headers[0], body

    def call_sym(self, req, **kwargs):
        return self.call_app(req, app=self.sym, **kwargs)


class TestSymlinkMiddleware(TestSymlinkMiddlewareBase):

    def test_symlink_info(self):
        swift_info = get_swift_info()
        self.assertEqual(swift_info['symlink'], {
            'symloop_max': 2,
            'static_links': True,
        })

    def test_symlink_simple_put(self):
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={'X-Symlink-Target': 'c1/o'},
                            body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '201 Created')
        method, path, hdrs = self.app.calls_with_headers[0]
        val = hdrs.get('X-Object-Sysmeta-Symlink-Target')
        self.assertEqual(val, 'c1/o')
        self.assertNotIn('X-Object-Sysmeta-Symlink-Target-Account', hdrs)
        val = hdrs.get('X-Object-Sysmeta-Container-Update-Override-Etag')
        self.assertEqual(val, '%s; symlink_target=c1/o' % MD5_OF_EMPTY_STRING)
        self.assertEqual('application/symlink', hdrs.get('Content-Type'))

    def test_symlink_simple_put_with_content_type(self):
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={'X-Symlink-Target': 'c1/o',
                                     'Content-Type': 'application/linkyfoo'},
                            body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '201 Created')
        method, path, hdrs = self.app.calls_with_headers[0]
        val = hdrs.get('X-Object-Sysmeta-Symlink-Target')
        self.assertEqual(val, 'c1/o')
        self.assertNotIn('X-Object-Sysmeta-Symlink-Target-Account', hdrs)
        val = hdrs.get('X-Object-Sysmeta-Container-Update-Override-Etag')
        self.assertEqual(val, '%s; symlink_target=c1/o' % MD5_OF_EMPTY_STRING)
        self.assertEqual('application/linkyfoo', hdrs.get('Content-Type'))

    def test_symlink_simple_put_with_etag(self):
        self.app.register('HEAD', '/v1/a/c1/o', swob.HTTPOk, {
            'Etag': 'tgt-etag', 'Content-Length': 42,
            'Content-Type': 'application/foo'})
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={
                                'X-Symlink-Target': 'c1/o',
                                'X-Symlink-Target-Etag': 'tgt-etag',
                            }, body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '201 Created')
        method, path, hdrs = self.app.calls_with_headers[1]
        val = hdrs.get('X-Object-Sysmeta-Symlink-Target')
        self.assertEqual(val, 'c1/o')
        self.assertNotIn('X-Object-Sysmeta-Symlink-Target-Account', hdrs)
        val = hdrs.get('X-Object-Sysmeta-Container-Update-Override-Etag')
        self.assertEqual(val, '%s; symlink_target=c1/o; '
                         'symlink_target_etag=tgt-etag; '
                         'symlink_target_bytes=42' % MD5_OF_EMPTY_STRING)
        self.assertEqual([
            ('HEAD', '/v1/a/c1/o'),
            ('PUT', '/v1/a/c/symlink'),
        ], self.app.calls)
        self.assertEqual('application/foo',
                         self.app.call_list[-1].headers['Content-Type'])

    def test_symlink_simple_put_with_quoted_etag(self):
        self.app.register('HEAD', '/v1/a/c1/o', swob.HTTPOk, {
            'Etag': 'tgt-etag', 'Content-Length': 42,
            'Content-Type': 'application/foo'})
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={
                                'X-Symlink-Target': 'c1/o',
                                'X-Symlink-Target-Etag': '"tgt-etag"',
                            }, body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '201 Created')
        method, path, hdrs = self.app.calls_with_headers[1]
        val = hdrs.get('X-Object-Sysmeta-Symlink-Target')
        self.assertEqual(val, 'c1/o')
        self.assertNotIn('X-Object-Sysmeta-Symlink-Target-Account', hdrs)
        val = hdrs.get('X-Object-Sysmeta-Container-Update-Override-Etag')
        self.assertEqual(val, '%s; symlink_target=c1/o; '
                         'symlink_target_etag=tgt-etag; '
                         'symlink_target_bytes=42' % MD5_OF_EMPTY_STRING)
        self.assertEqual([
            ('HEAD', '/v1/a/c1/o'),
            ('PUT', '/v1/a/c/symlink'),
        ], self.app.calls)
        self.assertEqual('application/foo',
                         self.app.call_list[-1].headers['Content-Type'])

    def test_symlink_simple_put_with_etag_target_missing_content_type(self):
        self.app.register('HEAD', '/v1/a/c1/o', swob.HTTPOk, {
            'Etag': 'tgt-etag', 'Content-Length': 42})
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={
                                'X-Symlink-Target': 'c1/o',
                                'X-Symlink-Target-Etag': 'tgt-etag',
                            }, body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '201 Created')
        method, path, hdrs = self.app.calls_with_headers[1]
        val = hdrs.get('X-Object-Sysmeta-Symlink-Target')
        self.assertEqual(val, 'c1/o')
        self.assertNotIn('X-Object-Sysmeta-Symlink-Target-Account', hdrs)
        val = hdrs.get('X-Object-Sysmeta-Container-Update-Override-Etag')
        self.assertEqual(val, '%s; symlink_target=c1/o; '
                         'symlink_target_etag=tgt-etag; '
                         'symlink_target_bytes=42' % MD5_OF_EMPTY_STRING)
        self.assertEqual([
            ('HEAD', '/v1/a/c1/o'),
            ('PUT', '/v1/a/c/symlink'),
        ], self.app.calls)
        # N.B. the ObjectController would call _update_content_type on PUT
        # regardless, but you actually can't get a HEAD response without swob
        # setting a Content-Type
        self.assertEqual('text/html; charset=UTF-8',
                         self.app.call_list[-1].headers['Content-Type'])

    def test_symlink_simple_put_with_etag_explicit_content_type(self):
        self.app.register('HEAD', '/v1/a/c1/o', swob.HTTPOk, {
            'Etag': 'tgt-etag', 'Content-Length': 42,
            'Content-Type': 'application/foo'})
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={
                                'X-Symlink-Target': 'c1/o',
                                'X-Symlink-Target-Etag': 'tgt-etag',
                                'Content-Type': 'application/bar',
                            }, body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '201 Created')
        method, path, hdrs = self.app.calls_with_headers[1]
        val = hdrs.get('X-Object-Sysmeta-Symlink-Target')
        self.assertEqual(val, 'c1/o')
        self.assertNotIn('X-Object-Sysmeta-Symlink-Target-Account', hdrs)
        val = hdrs.get('X-Object-Sysmeta-Container-Update-Override-Etag')
        self.assertEqual(val, '%s; symlink_target=c1/o; '
                         'symlink_target_etag=tgt-etag; '
                         'symlink_target_bytes=42' % MD5_OF_EMPTY_STRING)
        self.assertEqual([
            ('HEAD', '/v1/a/c1/o'),
            ('PUT', '/v1/a/c/symlink'),
        ], self.app.calls)
        self.assertEqual('application/bar',
                         self.app.call_list[-1].headers['Content-Type'])

    def test_symlink_simple_put_with_unmatched_etag(self):
        self.app.register('HEAD', '/v1/a/c1/o', swob.HTTPOk, {
            'Etag': 'tgt-etag', 'Content-Length': 42})
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={
                                'X-Symlink-Target': 'c1/o',
                                'X-Symlink-Target-Etag': 'not-tgt-etag',
                            }, body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '409 Conflict')
        self.assertIn(('Content-Location', '/v1/a/c1/o'), headers)
        self.assertEqual(body, b"Object Etag 'tgt-etag' does not match "
                         b"X-Symlink-Target-Etag header 'not-tgt-etag'")

    def test_symlink_simple_put_to_non_existing_object(self):
        self.app.register('HEAD', '/v1/a/c1/o', swob.HTTPNotFound, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={
                                'X-Symlink-Target': 'c1/o',
                                'X-Symlink-Target-Etag': 'not-tgt-etag',
                            }, body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '409 Conflict')
        self.assertIn(('Content-Location', '/v1/a/c1/o'), headers)
        self.assertIn(b'does not exist', body)

    def test_symlink_simple_put_error(self):
        self.app.register('HEAD', '/v1/a/c1/o',
                          swob.HTTPInternalServerError, {}, 'bad news')
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={
                                'X-Symlink-Target': 'c1/o',
                                'X-Symlink-Target-Etag': 'not-tgt-etag',
                            }, body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '500 Internal Error')
        # this is a PUT response; so if we have a content-length...
        self.assertGreater(int(dict(headers)['Content-Length']), 0)
        # ... we better have a body!
        self.assertIn(b'Internal Error', body)

    def test_symlink_simple_put_to_non_existing_object_override(self):
        self.app.register('HEAD', '/v1/a/c1/o', swob.HTTPNotFound, {})
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={
                                'X-Symlink-Target': 'c1/o',
                                'X-Symlink-Target-Etag': 'some-tgt-etag',
                                # this header isn't normally sent with PUT
                                'X-Symlink-Target-Bytes': '13',
                            }, body='')
        # this can be set in container_sync
        req.environ['swift.symlink_override'] = True
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '201 Created')

    def test_symlink_put_with_prevalidated_etag(self):
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT', headers={
            'X-Symlink-Target': 'c1/o',
            'X-Object-Sysmeta-Symlink-Target-Etag': 'tgt-etag',
            'X-Object-Sysmeta-Symlink-Target-Bytes': '13',
            'Content-Type': 'application/foo',
        }, body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '201 Created')

        self.assertEqual([
            # N.B. no HEAD!
            ('PUT', '/v1/a/c/symlink'),
        ], self.app.calls)
        self.assertEqual('application/foo',
                         self.app.call_list[-1].headers['Content-Type'])

        method, path, hdrs = self.app.calls_with_headers[0]
        val = hdrs.get('X-Object-Sysmeta-Symlink-Target')
        self.assertEqual(val, 'c1/o')
        self.assertNotIn('X-Object-Sysmeta-Symlink-Target-Account', hdrs)
        val = hdrs.get('X-Object-Sysmeta-Container-Update-Override-Etag')
        self.assertEqual(val, '%s; symlink_target=c1/o; '
                         'symlink_target_etag=tgt-etag; '
                         'symlink_target_bytes=13' % MD5_OF_EMPTY_STRING)

    def test_symlink_put_with_prevalidated_etag_sysmeta_incomplete(self):
        req = Request.blank('/v1/a/c/symlink', method='PUT', headers={
            'X-Symlink-Target': 'c1/o',
            'X-Object-Sysmeta-Symlink-Target-Etag': 'tgt-etag',
        }, body='')
        with self.assertRaises(KeyError) as cm:
            self.call_sym(req)
        self.assertEqual(cm.exception.args[0], swob.header_to_environ_key(
            'X-Object-Sysmeta-Symlink-Target-Bytes'))

    def test_symlink_chunked_put(self):
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={'X-Symlink-Target': 'c1/o'},
                            environ={'wsgi.input': io.BytesIO(b'')})
        self.assertIsNone(req.content_length)  # sanity
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '201 Created')
        method, path, hdrs = self.app.calls_with_headers[0]
        val = hdrs.get('X-Object-Sysmeta-Symlink-Target')
        self.assertEqual(val, 'c1/o')
        self.assertNotIn('X-Object-Sysmeta-Symlink-Target-Account', hdrs)
        val = hdrs.get('X-Object-Sysmeta-Container-Update-Override-Etag')
        self.assertEqual(val, '%s; symlink_target=c1/o' % MD5_OF_EMPTY_STRING)

    def test_symlink_chunked_put_error(self):
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={'X-Symlink-Target': 'c1/o'},
                            environ={'wsgi.input':
                                     io.BytesIO(b'this has a body')})
        self.assertIsNone(req.content_length)  # sanity
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '400 Bad Request')

    def test_symlink_put_different_account(self):
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={'X-Symlink-Target': 'c1/o',
                                     'X-Symlink-Target-Account': 'a1'},
                            body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '201 Created')
        method, path, hdrs = self.app.calls_with_headers[0]
        val = hdrs.get('X-Object-Sysmeta-Symlink-Target')
        self.assertEqual(val, 'c1/o')
        self.assertEqual(hdrs.get('X-Object-Sysmeta-Symlink-Target-Account'),
                         'a1')

    def test_symlink_put_leading_slash(self):
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={'X-Symlink-Target': '/c1/o'},
                            body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '412 Precondition Failed')
        self.assertEqual(body, b"X-Symlink-Target header must be of "
                               b"the form <container name>/<object name>")

    def test_symlink_put_non_zero_length(self):
        req = Request.blank('/v1/a/c/symlink', method='PUT', body='req_body',
                            headers={'X-Symlink-Target': 'c1/o'})
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '400 Bad Request')
        self.assertEqual(body, b'Symlink requests require a zero byte body')

    def test_symlink_put_bad_object_header(self):
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={'X-Symlink-Target': 'o'},
                            body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, "412 Precondition Failed")
        self.assertEqual(body, b"X-Symlink-Target header must be of "
                               b"the form <container name>/<object name>")

    def test_symlink_put_bad_account_header(self):
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={'X-Symlink-Target': 'c1/o',
                                     'X-Symlink-Target-Account': 'a1/c1'},
                            body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, "412 Precondition Failed")
        self.assertEqual(body, b"Account name cannot contain slashes")

    def test_get_symlink(self):
        self.app.register('GET', '/v1/a/c/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Meta-Color': 'Red'})
        req = Request.blank('/v1/a/c/symlink?symlink=get', method='GET')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')
        self.assertIsInstance(headers, list)
        self.assertIn(('X-Symlink-Target', 'c1/o'), headers)
        self.assertNotIn('X-Symlink-Target-Account', dict(headers))
        self.assertIn(('X-Object-Meta-Color', 'Red'), headers)
        self.assertEqual(body, b'')
        # HEAD with same params find same registered GET
        req = Request.blank('/v1/a/c/symlink?symlink=get', method='HEAD')
        head_status, head_headers, head_body = self.call_sym(req)
        self.assertEqual(head_status, '200 OK')
        self.assertIn(('X-Symlink-Target', 'c1/o'), head_headers)
        self.assertNotIn('X-Symlink-Target-Account', dict(head_headers))
        self.assertIn(('X-Object-Meta-Color', 'Red'), head_headers)
        self.assertEqual(head_body, b'')

    def test_get_symlink_with_account(self):
        self.app.register('GET', '/v1/a/c/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Sysmeta-Symlink-Target-Account': 'a2'})
        req = Request.blank('/v1/a/c/symlink?symlink=get', method='GET')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')
        self.assertIn(('X-Symlink-Target', 'c1/o'), headers)
        self.assertIn(('X-Symlink-Target-Account', 'a2'), headers)

    def test_get_symlink_not_found(self):
        self.app.register('GET', '/v1/a/c/symlink', swob.HTTPNotFound, {})
        req = Request.blank('/v1/a/c/symlink', method='GET')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '404 Not Found')
        self.assertNotIn('Content-Location', dict(headers))

    def test_get_target_object(self):
        self.app.register('GET', '/v1/a/c/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Sysmeta-Symlink-Target-Account': 'a2'})
        self.app.register('GET', '/v1/a2/c1/o', swob.HTTPOk, {}, 'resp_body')
        req_headers = {'X-Newest': 'True', 'X-Backend-Something': 'future'}
        req = Request.blank('/v1/a/c/symlink', method='GET',
                            headers=req_headers)
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')
        self.assertEqual(body, b'resp_body')
        self.assertNotIn('X-Symlink-Target', dict(headers))
        self.assertNotIn('X-Symlink-Target-Account', dict(headers))
        self.assertIn(('Content-Location', '/v1/a2/c1/o'), headers)
        calls = self.app.call_list
        req_headers.update({
            'Host': 'localhost:80',
            'X-Backend-Ignore-Range-If-Metadata-Present':
            'x-object-sysmeta-symlink-target',
        })
        self.assertEqual(req_headers, calls[0].headers)
        req_headers['User-Agent'] = 'Swift'
        self.assertEqual(req_headers, calls[1].headers)
        self.assertFalse(calls[2:])
        self.assertFalse(self.app.unread_requests)

    def test_get_target_object_not_found(self):
        self.app.register('GET', '/v1/a/c/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Sysmeta-Symlink-Target-account': 'a2'})
        self.app.register('GET', '/v1/a2/c1/o', swob.HTTPNotFound, {}, '')
        req = Request.blank('/v1/a/c/symlink', method='GET')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '404 Not Found')
        self.assertEqual(body, b'')
        self.assertNotIn('X-Symlink-Target', dict(headers))
        self.assertNotIn('X-Symlink-Target-Account', dict(headers))
        self.assertIn(('Content-Location', '/v1/a2/c1/o'), headers)
        self.assertFalse(self.app.unread_requests)

    def test_get_target_object_range_not_satisfiable(self):
        self.app.register('GET', '/v1/a/c/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Sysmeta-Symlink-Target-Account': 'a2'})
        self.app.register('GET', '/v1/a2/c1/o',
                          swob.HTTPRequestedRangeNotSatisfiable, {}, '')
        req = Request.blank('/v1/a/c/symlink', method='GET',
                            headers={'Range': 'bytes=1-2'})
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '416 Requested Range Not Satisfiable')
        self.assertEqual(
            body, b'<html><h1>Requested Range Not Satisfiable</h1>'
                  b'<p>The Range requested is not available.</p></html>')
        self.assertNotIn('X-Symlink-Target', dict(headers))
        self.assertNotIn('X-Symlink-Target-Account', dict(headers))
        self.assertIn(('Content-Location', '/v1/a2/c1/o'), headers)
        self.assertFalse(self.app.unread_requests)

    def test_get_ec_symlink_range_unsatisfiable_can_redirect_to_target(self):
        self.app.register('GET', '/v1/a/c/symlink',
                          swob.HTTPRequestedRangeNotSatisfiable,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Sysmeta-Symlink-Target-Account': 'a2'})
        self.app.register('GET', '/v1/a2/c1/o', swob.HTTPOk,
                          {'Content-Range': 'bytes 1-2/10'}, 'es')
        req = Request.blank('/v1/a/c/symlink', method='GET',
                            headers={'Range': 'bytes=1-2'})
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')
        self.assertEqual(body, b'es')
        self.assertNotIn('X-Symlink-Target', dict(headers))
        self.assertNotIn('X-Symlink-Target-Account', dict(headers))
        self.assertIn(('Content-Location', '/v1/a2/c1/o'), headers)
        self.assertIn(('Content-Range', 'bytes 1-2/10'), headers)

    def test_get_non_symlink(self):
        # this is not symlink object
        self.app.register('GET', '/v1/a/c/obj', swob.HTTPOk, {}, 'resp_body')
        req = Request.blank('/v1/a/c/obj', method='GET')

        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')
        self.assertEqual(body, b'resp_body')

        # Assert special headers for symlink are not in response
        self.assertNotIn('X-Symlink-Target', dict(headers))
        self.assertNotIn('X-Symlink-Target-Account', dict(headers))
        self.assertNotIn('Content-Location', dict(headers))

    def test_get_static_link_mismatched_etag(self):
        self.app.register('GET', '/v1/a/c/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Sysmeta-Symlink-Target-Etag': 'the-etag'})
        # apparently target object was overwritten
        self.app.register('GET', '/v1/a/c1/o', swob.HTTPOk,
                          {'ETag': 'not-the-etag'}, 'resp_body')
        req = Request.blank('/v1/a/c/symlink', method='GET')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '409 Conflict')
        self.assertEqual(body, b"Object Etag 'not-the-etag' does not "
                         b"match X-Symlink-Target-Etag header 'the-etag'")

    def test_get_static_link_to_symlink(self):
        self.app.register('GET', '/v1/a/c/static_link', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c/symlink',
                           'X-Object-Sysmeta-Symlink-Target-Etag': 'the-etag'})
        self.app.register('GET', '/v1/a/c/symlink', swob.HTTPOk,
                          {'ETag': 'the-etag',
                           'X-Object-Sysmeta-Symlink-Target': 'c1/o'})
        self.app.register('GET', '/v1/a/c1/o', swob.HTTPOk,
                          {'ETag': 'not-the-etag'}, 'resp_body')
        req = Request.blank('/v1/a/c/static_link', method='GET')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')

    def test_get_static_link_to_symlink_fails(self):
        self.app.register('GET', '/v1/a/c/static_link', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c/symlink',
                           'X-Object-Sysmeta-Symlink-Target-Etag': 'the-etag'})
        self.app.register('GET', '/v1/a/c/symlink', swob.HTTPOk,
                          {'ETag': 'not-the-etag',
                           'X-Object-Sysmeta-Symlink-Target': 'c1/o'})
        req = Request.blank('/v1/a/c/static_link', method='GET')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '409 Conflict')
        self.assertEqual(body, b"X-Symlink-Target-Etag headers do not match")

    def put_static_link_to_symlink(self):
        self.app.register('HEAD', '/v1/a/c/symlink', swob.HTTPOk,
                          {'ETag': 'symlink-etag',
                           'X-Object-Sysmeta-Symlink-Target': 'c/o',
                           'Content-Type': 'application/symlink'})
        self.app.register('HEAD', '/v1/a/c/o', swob.HTTPOk,
                          {'ETag': 'tgt-etag',
                           'Content-Type': 'application/data'}, 'resp_body')
        self.app.register('PUT', '/v1/a/c/static_link', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/static_link', method='PUT',
                            headers={
                                'X-Symlink-Target': 'c/symlink',
                                'X-Symlink-Target-Etag': 'symlink-etag',
                            }, body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '201 Created')
        self.assertEqual([], self.app.calls)
        self.assertEqual('application/data',
                         self.app.call_list[-1].headers['Content-Type'])

    def test_head_symlink(self):
        self.app.register('HEAD', '/v1/a/c/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Meta-Color': 'Red'})
        req = Request.blank('/v1/a/c/symlink?symlink=get', method='HEAD')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')
        self.assertIn(('X-Symlink-Target', 'c1/o'), headers)
        self.assertNotIn('X-Symlink-Target-Account', dict(headers))
        self.assertIn(('X-Object-Meta-Color', 'Red'), headers)

    def test_head_symlink_with_account(self):
        self.app.register('HEAD', '/v1/a/c/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Sysmeta-Symlink-Target-Account': 'a2',
                           'X-Object-Meta-Color': 'Red'})
        req = Request.blank('/v1/a/c/symlink?symlink=get', method='HEAD')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')
        self.assertIn(('X-Symlink-Target', 'c1/o'), headers)
        self.assertIn(('X-Symlink-Target-Account', 'a2'), headers)
        self.assertIn(('X-Object-Meta-Color', 'Red'), headers)

    def test_head_target_object(self):
        # this test is also validating that the symlink metadata is not
        # returned, but the target object metadata does return
        self.app.register('HEAD', '/v1/a/c/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Sysmeta-Symlink-Target-Account': 'a2',
                           'X-Object-Meta-Color': 'Red'})
        self.app.register('HEAD', '/v1/a2/c1/o', swob.HTTPOk,
                          {'X-Object-Meta-Color': 'Green'})
        req_headers = {'X-Newest': 'True', 'X-Backend-Something': 'future'}
        req = Request.blank('/v1/a/c/symlink', method='HEAD',
                            headers=req_headers)
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')
        self.assertNotIn('X-Symlink-Target', dict(headers))
        self.assertNotIn('X-Symlink-Target-Account', dict(headers))
        self.assertNotIn(('X-Object-Meta-Color', 'Red'), headers)
        self.assertIn(('X-Object-Meta-Color', 'Green'), headers)
        self.assertIn(('Content-Location', '/v1/a2/c1/o'), headers)
        calls = self.app.call_list
        req_headers.update({
            'Host': 'localhost:80',
            'X-Backend-Ignore-Range-If-Metadata-Present':
            'x-object-sysmeta-symlink-target',
        })
        self.assertEqual(req_headers, calls[0].headers)
        req_headers['User-Agent'] = 'Swift'
        self.assertEqual(req_headers, calls[1].headers)
        self.assertFalse(calls[2:])

    def test_get_symlink_to_reserved_object(self):
        cont = get_reserved_name('versioned')
        obj = get_reserved_name('symlink', '9999998765.99999')
        symlink_target = "%s/%s" % (cont, obj)
        version_path = '/v1/a/%s' % symlink_target
        self.app.register('GET', '/v1/a/versioned/symlink', swob.HTTPOk, {
            symlink.TGT_OBJ_SYSMETA_SYMLINK_HDR: symlink_target,
            symlink.ALLOW_RESERVED_NAMES: 'true',
            'x-object-sysmeta-symlink-target-etag': MD5_OF_EMPTY_STRING,
            'x-object-sysmeta-symlink-target-bytes': '0',
        })
        self.app.register('GET', version_path, swob.HTTPOk, {})
        req = Request.blank('/v1/a/versioned/symlink', headers={
            'Range': 'foo', 'If-Match': 'bar'})
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')
        self.assertIn(('Content-Location', version_path), headers)
        self.assertEqual(len(self.authorized), 1)
        self.assertNotIn('X-Backend-Allow-Reserved-Names',
                         self.app.calls_with_headers[0])
        call_headers = self.app.call_list[1].headers
        self.assertEqual('true', call_headers[
            'X-Backend-Allow-Reserved-Names'])
        self.assertEqual('foo', call_headers['Range'])
        self.assertEqual('bar', call_headers['If-Match'])

    def test_get_symlink_to_reserved_symlink(self):
        cont = get_reserved_name('versioned')
        obj = get_reserved_name('symlink', '9999998765.99999')
        symlink_target = "%s/%s" % (cont, obj)
        version_path = '/v1/a/%s' % symlink_target
        self.app.register('GET', '/v1/a/versioned/symlink', swob.HTTPOk, {
            symlink.TGT_OBJ_SYSMETA_SYMLINK_HDR: symlink_target,
            symlink.ALLOW_RESERVED_NAMES: 'true',
            'x-object-sysmeta-symlink-target-etag': MD5_OF_EMPTY_STRING,
            'x-object-sysmeta-symlink-target-bytes': '0',
        })
        self.app.register('GET', version_path, swob.HTTPOk, {
            symlink.TGT_OBJ_SYSMETA_SYMLINK_HDR: 'unversioned/obj',
            'ETag': MD5_OF_EMPTY_STRING,
        })
        self.app.register('GET', '/v1/a/unversioned/obj', swob.HTTPOk, {
        })
        req = Request.blank('/v1/a/versioned/symlink')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')
        self.assertIn(('Content-Location', '/v1/a/unversioned/obj'), headers)
        self.assertEqual(len(self.authorized), 2)

    def test_symlink_too_deep(self):
        self.app.register('GET', '/v1/a/c/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c/sym1'})
        self.app.register('GET', '/v1/a/c/sym1', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c/sym2'})
        self.app.register('GET', '/v1/a/c/sym2', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c/o'})
        req = Request.blank('/v1/a/c/symlink', method='HEAD')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '409 Conflict')
        self.assertEqual(body, b'')
        req = Request.blank('/v1/a/c/symlink')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '409 Conflict')
        self.assertEqual(body, b'Too many levels of symbolic links, '
                         b'maximum allowed is 2')

    def test_symlink_change_symloopmax(self):
        # similar test to test_symlink_too_deep, but now changed the limit to 3
        self.sym = symlink.filter_factory({
            'symloop_max': '3',
        })(self.app)
        self.sym.logger = self.app.logger
        self.app.register('HEAD', '/v1/a/c/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c/sym1'})
        self.app.register('HEAD', '/v1/a/c/sym1', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c/sym2'})
        self.app.register('HEAD', '/v1/a/c/sym2', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c/o',
                           'X-Object-Meta-Color': 'Red'})
        self.app.register('HEAD', '/v1/a/c/o', swob.HTTPOk,
                          {'X-Object-Meta-Color': 'Green'})
        req = Request.blank('/v1/a/c/symlink', method='HEAD')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')

        # assert that the correct metadata was returned
        self.assertNotIn(('X-Object-Meta-Color', 'Red'), headers)
        self.assertIn(('X-Object-Meta-Color', 'Green'), headers)

    def test_sym_to_sym_to_target(self):
        # this test is also validating that the symlink metadata is not
        # returned, but the target object metadata does return
        self.app.register('HEAD', '/v1/a/c/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c/sym1',
                           'X-Object-Meta-Color': 'Red'})
        self.app.register('HEAD', '/v1/a/c/sym1', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Meta-Color': 'Yellow'})
        self.app.register('HEAD', '/v1/a/c1/o', swob.HTTPOk,
                          {'X-Object-Meta-Color': 'Green'})
        req = Request.blank('/v1/a/c/symlink', method='HEAD')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')
        self.assertNotIn(('X-Symlink-Target', 'c1/o'), headers)
        self.assertNotIn(('X-Symlink-Target-Account', 'a2'), headers)
        self.assertNotIn(('X-Object-Meta-Color', 'Red'), headers)
        self.assertNotIn(('X-Object-Meta-Color', 'Yellow'), headers)
        self.assertIn(('X-Object-Meta-Color', 'Green'), headers)
        self.assertIn(('Content-Location', '/v1/a/c1/o'), headers)

    def test_symlink_post(self):
        self.app.register('POST', '/v1/a/c/symlink', swob.HTTPAccepted,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o'})
        req = Request.blank('/v1/a/c/symlink', method='POST',
                            headers={'X-Object-Meta-Color': 'Red'})
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '307 Temporary Redirect')
        self.assertEqual(
            body,
            b'The requested POST was applied to a symlink. POST '
            b'directly to the target to apply requested metadata.')
        method, path, hdrs = self.app.calls_with_headers[0]
        val = hdrs.get('X-Object-Meta-Color')
        self.assertEqual(val, 'Red')

    def test_non_symlink_post(self):
        self.app.register('POST', '/v1/a/c/o', swob.HTTPAccepted, {})
        req = Request.blank('/v1/a/c/o', method='POST',
                            headers={'X-Object-Meta-Color': 'Red'})
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '202 Accepted')

    def test_set_symlink_POST_fail(self):
        # Setting a link with a POST request is not allowed
        req = Request.blank('/v1/a/c/o', method='POST',
                            headers={'X-Symlink-Target': 'c1/regular_obj'})
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '400 Bad Request')
        self.assertEqual(body, b"A PUT request is required to set a symlink "
                         b"target")

    def test_symlink_post_but_fail_at_server(self):
        self.app.register('POST', '/v1/a/c/o', swob.HTTPNotFound, {})
        req = Request.blank('/v1/a/c/o', method='POST',
                            headers={'X-Object-Meta-Color': 'Red'})
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '404 Not Found')

    def test_validate_and_prep_request_headers(self):
        def do_test(headers):
            req = Request.blank('/v1/a/c/o', method='PUT',
                                headers=headers)
            symlink._validate_and_prep_request_headers(req)

        # normal cases
        do_test({'X-Symlink-Target': 'c1/o1'})
        do_test({'X-Symlink-Target': 'c1/sub/o1'})
        do_test({'X-Symlink-Target': 'c1%2Fo1'})
        # specify account
        do_test({'X-Symlink-Target': 'c1/o1',
                 'X-Symlink-Target-Account': 'another'})
        # URL encoded is safe
        do_test({'X-Symlink-Target': 'c1%2Fo1'})
        # URL encoded + multibytes is also safe
        target = u'\u30b0\u30e9\u30d6\u30eb/\u30a2\u30ba\u30ec\u30f3'
        target = swob.bytes_to_wsgi(target.encode('utf8'))
        do_test({'X-Symlink-Target': target})
        do_test({'X-Symlink-Target': swob.wsgi_quote(target)})

        target = swob.bytes_to_wsgi(u'\u30b0\u30e9\u30d6\u30eb'.encode('utf8'))
        do_test(
            {'X-Symlink-Target': 'cont/obj',
             'X-Symlink-Target-Account': target})
        do_test(
            {'X-Symlink-Target': 'cont/obj',
             'X-Symlink-Target-Account': swob.wsgi_quote(target)})

    def test_validate_and_prep_request_headers_invalid_format(self):
        def do_test(headers, status, err_msg):
            req = Request.blank('/v1/a/c/o', method='PUT',
                                headers=headers)
            with self.assertRaises(swob.HTTPException) as cm:
                symlink._validate_and_prep_request_headers(req)

            self.assertEqual(cm.exception.status, status)
            self.assertEqual(cm.exception.body, err_msg)

        do_test({'X-Symlink-Target': '/c1/o1'},
                '412 Precondition Failed',
                b'X-Symlink-Target header must be of the '
                b'form <container name>/<object name>')
        do_test({'X-Symlink-Target': 'c1o1'},
                '412 Precondition Failed',
                b'X-Symlink-Target header must be of the '
                b'form <container name>/<object name>')
        do_test({'X-Symlink-Target': 'c1/o1',
                 'X-Symlink-Target-Account': '/another'},
                '412 Precondition Failed',
                b'Account name cannot contain slashes')
        do_test({'X-Symlink-Target': 'c1/o1',
                 'X-Symlink-Target-Account': 'an/other'},
                '412 Precondition Failed',
                b'Account name cannot contain slashes')
        # url encoded case
        do_test({'X-Symlink-Target': '%2Fc1%2Fo1'},
                '412 Precondition Failed',
                b'X-Symlink-Target header must be of the '
                b'form <container name>/<object name>')
        do_test({'X-Symlink-Target': 'c1/o1',
                 'X-Symlink-Target-Account': '%2Fanother'},
                '412 Precondition Failed',
                b'Account name cannot contain slashes')
        do_test({'X-Symlink-Target': 'c1/o1',
                 'X-Symlink-Target-Account': 'an%2Fother'},
                '412 Precondition Failed',
                b'Account name cannot contain slashes')
        # with multi-bytes
        target = u'/\u30b0\u30e9\u30d6\u30eb/\u30a2\u30ba\u30ec\u30f3'
        target = swob.bytes_to_wsgi(target.encode('utf8'))
        do_test(
            {'X-Symlink-Target': target},
            '412 Precondition Failed',
            b'X-Symlink-Target header must be of the '
            b'form <container name>/<object name>')
        do_test(
            {'X-Symlink-Target': swob.wsgi_quote(target)},
            '412 Precondition Failed',
            b'X-Symlink-Target header must be of the '
            b'form <container name>/<object name>')
        account = u'\u30b0\u30e9\u30d6\u30eb/\u30a2\u30ba\u30ec\u30f3'
        account = swob.bytes_to_wsgi(account.encode('utf8'))
        do_test(
            {'X-Symlink-Target': 'c/o',
             'X-Symlink-Target-Account': account},
            '412 Precondition Failed',
            b'Account name cannot contain slashes')
        do_test(
            {'X-Symlink-Target': 'c/o',
             'X-Symlink-Target-Account': swob.wsgi_quote(account)},
            '412 Precondition Failed',
            b'Account name cannot contain slashes')

    def test_validate_and_prep_request_headers_points_to_itself(self):
        req = Request.blank('/v1/a/c/o', method='PUT',
                            headers={'X-Symlink-Target': 'c/o'})
        with self.assertRaises(swob.HTTPException) as cm:
            symlink._validate_and_prep_request_headers(req)
        self.assertEqual(cm.exception.status, '400 Bad Request')
        self.assertEqual(cm.exception.body, b'Symlink cannot target itself')

        # Even if set account to itself, it will fail as well
        req = Request.blank('/v1/a/c/o', method='PUT',
                            headers={'X-Symlink-Target': 'c/o',
                                     'X-Symlink-Target-Account': 'a'})
        with self.assertRaises(swob.HTTPException) as cm:
            symlink._validate_and_prep_request_headers(req)
        self.assertEqual(cm.exception.status, '400 Bad Request')
        self.assertEqual(cm.exception.body, b'Symlink cannot target itself')

        # sanity, the case to another account is safe
        req = Request.blank('/v1/a/c/o', method='PUT',
                            headers={'X-Symlink-Target': 'c/o',
                                     'X-Symlink-Target-Account': 'a1'})
        symlink._validate_and_prep_request_headers(req)

    def test_symloop_max_config(self):
        self.app = FakeSwift()
        # sanity
        self.sym = symlink.filter_factory({
            'symloop_max': '1',
        })(self.app)
        self.assertEqual(self.sym.symloop_max, 1)
        # < 1 case will result in default
        self.sym = symlink.filter_factory({
            'symloop_max': '-1',
        })(self.app)
        self.assertEqual(self.sym.symloop_max, symlink.DEFAULT_SYMLOOP_MAX)


class SymlinkCopyingTestCase(TestSymlinkMiddlewareBase):
    # verify interaction of copy and symlink middlewares

    def setUp(self):
        self.app = FakeSwift()
        conf = {'symloop_max': '2'}
        self.sym = symlink.filter_factory(conf)(self.app)
        self.sym.logger = self.app.logger
        self.copy = copy.filter_factory({})(self.sym)

    def call_copy(self, req, **kwargs):
        return self.call_app(req, app=self.copy, **kwargs)

    def test_copy_symlink_target(self):
        req = Request.blank('/v1/a/src_cont/symlink', method='COPY',
                            headers={'Destination': 'tgt_cont/tgt_obj'})
        self._test_copy_symlink_target(req)
        req = Request.blank('/v1/a/tgt_cont/tgt_obj', method='PUT',
                            headers={'X-Copy-From': 'src_cont/symlink'})
        self._test_copy_symlink_target(req)

    def _test_copy_symlink_target(self, req):
        self.app.register('GET', '/v1/a/src_cont/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Sysmeta-Symlink-Target-Account': 'a2'})
        self.app.register('GET', '/v1/a2/c1/o', swob.HTTPOk, {}, 'resp_body')
        self.app.register('PUT', '/v1/a/tgt_cont/tgt_obj', swob.HTTPCreated,
                          {}, 'resp_body')
        status, headers, body = self.call_copy(req)
        method, path, hdrs = self.app.calls_with_headers[0]
        self.assertEqual(method, 'GET')
        self.assertEqual(path, '/v1/a/src_cont/symlink')
        self.assertEqual('/src_cont/symlink', hdrs.get('X-Copy-From'))
        method, path, hdrs = self.app.calls_with_headers[1]
        self.assertEqual(method, 'GET')
        self.assertEqual(path, '/v1/a2/c1/o')
        self.assertEqual('/src_cont/symlink', hdrs.get('X-Copy-From'))
        method, path, hdrs = self.app.calls_with_headers[2]
        self.assertEqual(method, 'PUT')
        val = hdrs.get('X-Object-Sysmeta-Symlink-Target')
        # this is raw object copy
        self.assertEqual(val, None)
        self.assertEqual(status, '201 Created')

    def test_copy_symlink(self):
        req = Request.blank(
            '/v1/a/src_cont/symlink?symlink=get', method='COPY',
            headers={'Destination': 'tgt_cont/tgt_obj'})
        self._test_copy_symlink(req)
        req = Request.blank(
            '/v1/a/tgt_cont/tgt_obj?symlink=get', method='PUT',
            headers={'X-Copy-From': 'src_cont/symlink'})
        self._test_copy_symlink(req)

    def _test_copy_symlink(self, req):
        self.app.register('GET', '/v1/a/src_cont/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Sysmeta-Symlink-Target-Account': 'a2'})
        self.app.register('PUT', '/v1/a/tgt_cont/tgt_obj', swob.HTTPCreated,
                          {'X-Symlink-Target': 'c1/o',
                           'X-Symlink-Target-Account': 'a2'})
        status, headers, body = self.call_copy(req)
        self.assertEqual(status, '201 Created')
        method, path, hdrs = self.app.calls_with_headers[0]
        self.assertEqual(method, 'GET')
        self.assertEqual(path, '/v1/a/src_cont/symlink?symlink=get')
        self.assertEqual('/src_cont/symlink', hdrs.get('X-Copy-From'))
        method, path, hdrs = self.app.calls_with_headers[1]
        val = hdrs.get('X-Object-Sysmeta-Symlink-Target')
        self.assertEqual(val, 'c1/o')
        self.assertEqual(
            hdrs.get('X-Object-Sysmeta-Symlink-Target-Account'), 'a2')

    def test_copy_symlink_new_target(self):
        req = Request.blank(
            '/v1/a/src_cont/symlink?symlink=get', method='COPY',
            headers={'Destination': 'tgt_cont/tgt_obj',
                     'X-Symlink-Target': 'new_cont/new_obj',
                     'X-Symlink-Target-Account': 'new_acct'})
        self._test_copy_symlink_new_target(req)
        req = Request.blank(
            '/v1/a/tgt_cont/tgt_obj?symlink=get', method='PUT',
            headers={'X-Copy-From': 'src_cont/symlink',
                     'X-Symlink-Target': 'new_cont/new_obj',
                     'X-Symlink-Target-Account': 'new_acct'})
        self._test_copy_symlink_new_target(req)

    def _test_copy_symlink_new_target(self, req):
        self.app.register('GET', '/v1/a/src_cont/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Sysmeta-Symlink-Target-Account': 'a2'})
        self.app.register('PUT', '/v1/a/tgt_cont/tgt_obj', swob.HTTPCreated,
                          {'X-Symlink-Target': 'c1/o',
                           'X-Symlink-Target-Account': 'a2'})
        status, headers, body = self.call_copy(req)
        self.assertEqual(status, '201 Created')
        method, path, hdrs = self.app.calls_with_headers[0]
        self.assertEqual(method, 'GET')
        self.assertEqual(path, '/v1/a/src_cont/symlink?symlink=get')
        self.assertEqual('/src_cont/symlink', hdrs.get('X-Copy-From'))
        method, path, hdrs = self.app.calls_with_headers[1]
        self.assertEqual(method, 'PUT')
        self.assertEqual(path, '/v1/a/tgt_cont/tgt_obj?symlink=get')
        val = hdrs.get('X-Object-Sysmeta-Symlink-Target')
        self.assertEqual(val, 'new_cont/new_obj')
        self.assertEqual(hdrs.get('X-Object-Sysmeta-Symlink-Target-Account'),
                         'new_acct')

    def test_copy_symlink_with_slo_query(self):
        req = Request.blank(
            '/v1/a/src_cont/symlink?multipart-manifest=get&symlink=get',
            method='COPY', headers={'Destination': 'tgt_cont/tgt_obj'})
        self._test_copy_symlink_with_slo_query(req)
        req = Request.blank(
            '/v1/a/tgt_cont/tgt_obj?multipart-manifest=get&symlink=get',
            method='PUT', headers={'X-Copy-From': 'src_cont/symlink'})
        self._test_copy_symlink_with_slo_query(req)

    def _test_copy_symlink_with_slo_query(self, req):
        self.app.register('GET', '/v1/a/src_cont/symlink', swob.HTTPOk,
                          {'X-Object-Sysmeta-Symlink-Target': 'c1/o',
                           'X-Object-Sysmeta-Symlink-Target-Account': 'a2'})
        self.app.register('PUT', '/v1/a/tgt_cont/tgt_obj', swob.HTTPCreated,
                          {'X-Symlink-Target': 'c1/o',
                           'X-Symlink-Target-Account': 'a2'})
        status, headers, body = self.call_copy(req)
        self.assertEqual(status, '201 Created')
        method, path, hdrs = self.app.calls_with_headers[0]
        self.assertEqual(method, 'GET')
        path, query = path.split('?')
        query_dict = parse_qs(query)
        self.assertEqual(
            path, '/v1/a/src_cont/symlink')
        self.assertEqual(
            query_dict,
            {'multipart-manifest': ['get'], 'symlink': ['get'],
             'format': ['raw']})
        self.assertEqual('/src_cont/symlink', hdrs.get('X-Copy-From'))
        method, path, hdrs = self.app.calls_with_headers[1]
        val = hdrs.get('X-Object-Sysmeta-Symlink-Target')
        self.assertEqual(val, 'c1/o')
        self.assertEqual(
            hdrs.get('X-Object-Sysmeta-Symlink-Target-Account'), 'a2')

    def test_static_link_to_new_slo_manifest(self):
        self.app.register('HEAD', '/v1/a/c1/o', swob.HTTPOk, {
            'X-Static-Large-Object': 'True',
            'Etag': 'manifest-etag',
            'X-Object-Sysmeta-Slo-Size': '1048576',
            'X-Object-Sysmeta-Slo-Etag': 'this-is-not-used',
            'Content-Length': 42,
            'Content-Type': 'application/big-data',
            'X-Object-Sysmeta-Container-Update-Override-Etag':
            '956859738870e5ca6aa17eeda58e4df0; '
            'slo_etag=71e938d37c1d06dc634dd24660255a88',

        })
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={
                                'X-Symlink-Target': 'c1/o',
                                'X-Symlink-Target-Etag': 'manifest-etag',
                            }, body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '201 Created')
        self.assertEqual([
            ('HEAD', '/v1/a/c1/o'),
            ('PUT', '/v1/a/c/symlink'),
        ], self.app.calls)
        method, path, hdrs = self.app.calls_with_headers[-1]
        self.assertEqual('application/big-data', hdrs['Content-Type'])
        self.assertEqual(hdrs['X-Object-Sysmeta-Symlink-Target'], 'c1/o')
        self.assertEqual(hdrs['X-Object-Sysmeta-Symlink-Target-Etag'],
                         'manifest-etag')
        self.assertEqual(hdrs['X-Object-Sysmeta-Symlink-Target-Bytes'],
                         '1048576')
        self.assertEqual(
            hdrs['X-Object-Sysmeta-Container-Update-Override-Etag'],
            'd41d8cd98f00b204e9800998ecf8427e; '
            'slo_etag=71e938d37c1d06dc634dd24660255a88; '
            'symlink_target=c1/o; '
            'symlink_target_etag=manifest-etag; '
            'symlink_target_bytes=1048576')

    def test_static_link_to_old_slo_manifest(self):
        self.app.register('HEAD', '/v1/a/c1/o', swob.HTTPOk, {
            'X-Static-Large-Object': 'True',
            'Etag': 'manifest-etag',
            'X-Object-Sysmeta-Slo-Size': '1048576',
            'X-Object-Sysmeta-Slo-Etag': '71e938d37c1d06dc634dd24660255a88',
            'Content-Length': 42,
            'Content-Type': 'application/big-data',

        })
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={
                                'X-Symlink-Target': 'c1/o',
                                'X-Symlink-Target-Etag': 'manifest-etag',
                            }, body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '201 Created')
        self.assertEqual([
            ('HEAD', '/v1/a/c1/o'),
            ('PUT', '/v1/a/c/symlink'),
        ], self.app.calls)
        method, path, hdrs = self.app.calls_with_headers[-1]
        self.assertEqual('application/big-data', hdrs['Content-Type'])
        self.assertEqual(hdrs['X-Object-Sysmeta-Symlink-Target'], 'c1/o')
        self.assertEqual(hdrs['X-Object-Sysmeta-Symlink-Target-Etag'],
                         'manifest-etag')
        self.assertEqual(hdrs['X-Object-Sysmeta-Symlink-Target-Bytes'],
                         '1048576')
        self.assertEqual(
            hdrs['X-Object-Sysmeta-Container-Update-Override-Etag'],
            'd41d8cd98f00b204e9800998ecf8427e; '
            'slo_etag=71e938d37c1d06dc634dd24660255a88; '
            'symlink_target=c1/o; '
            'symlink_target_etag=manifest-etag; '
            'symlink_target_bytes=1048576')

    def test_static_link_to_really_old_slo_manifest(self):
        self.app.register('HEAD', '/v1/a/c1/o', swob.HTTPOk, {
            'X-Static-Large-Object': 'True',
            'Etag': 'manifest-etag',
            'Content-Length': 42,
            'Content-Type': 'application/big-data',

        })
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={
                                'X-Symlink-Target': 'c1/o',
                                'X-Symlink-Target-Etag': 'manifest-etag',
                            }, body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '201 Created')
        self.assertEqual([
            ('HEAD', '/v1/a/c1/o'),
            ('PUT', '/v1/a/c/symlink'),
        ], self.app.calls)
        method, path, hdrs = self.app.calls_with_headers[-1]
        self.assertEqual('application/big-data', hdrs['Content-Type'])
        self.assertEqual(hdrs['X-Object-Sysmeta-Symlink-Target'], 'c1/o')
        self.assertEqual(hdrs['X-Object-Sysmeta-Symlink-Target-Etag'],
                         'manifest-etag')
        # symlink m/w is doing a HEAD, it's not going to going to read the
        # manifest body and sum up the bytes - so we just use manifest size
        self.assertEqual(hdrs['X-Object-Sysmeta-Symlink-Target-Bytes'],
                         '42')
        # no slo_etag, and target_bytes is manifest
        self.assertEqual(
            hdrs['X-Object-Sysmeta-Container-Update-Override-Etag'],
            'd41d8cd98f00b204e9800998ecf8427e; '
            'symlink_target=c1/o; '
            'symlink_target_etag=manifest-etag; '
            'symlink_target_bytes=42')

    def test_static_link_to_slo_manifest_slo_etag(self):
        self.app.register('HEAD', '/v1/a/c1/o', swob.HTTPOk, {
            'Etag': 'manifest-etag',
            'X-Object-Sysmeta-Slo-Etag': 'slo-etag',
            'Content-Length': 42,
        })
        self.app.register('PUT', '/v1/a/c/symlink', swob.HTTPCreated, {})
        # unquoted slo-etag doesn't match
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={
                                'X-Symlink-Target': 'c1/o',
                                'X-Symlink-Target-Etag': 'slo-etag',
                            }, body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '409 Conflict')
        # the quoted slo-etag is tolerated, but still doesn't match
        req = Request.blank('/v1/a/c/symlink', method='PUT',
                            headers={
                                'X-Symlink-Target': 'c1/o',
                                'X-Symlink-Target-Etag': '"slo-etag"',
                            }, body='')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '409 Conflict')


class SymlinkVersioningTestCase(TestSymlinkMiddlewareBase):
    # verify interaction of versioned_writes and symlink middlewares

    def setUp(self):
        self.app = FakeSwift()
        conf = {'symloop_max': '2'}
        self.sym = symlink.filter_factory(conf)(self.app)
        self.sym.logger = self.app.logger
        vw_conf = {'allow_versioned_writes': 'true'}
        self.vw = versioned_writes.filter_factory(vw_conf)(self.sym)

    def call_vw(self, req, **kwargs):
        return self.call_app(req, app=self.vw, **kwargs)

    def assertRequestEqual(self, req, other):
        self.assertEqual(req.method, other.method)
        self.assertEqual(req.path, other.path)

    def test_new_symlink_version_success(self):
        self.app.register(
            'PUT', '/v1/a/c/symlink', swob.HTTPCreated,
            {'X-Symlink-Target': 'new_cont/new_tgt',
             'X-Symlink-Target-Account': 'a'}, None)
        self.app.register(
            'GET', '/v1/a/c/symlink', swob.HTTPOk,
            {'last-modified': 'Thu, 1 Jan 1970 00:00:01 GMT',
             'X-Object-Sysmeta-Symlink-Target': 'old_cont/old_tgt',
             'X-Object-Sysmeta-Symlink-Target-Account': 'a'},
            '')
        self.app.register(
            'PUT', '/v1/a/ver_cont/007symlink/0000000001.00000',
            swob.HTTPCreated,
            {'X-Symlink-Target': 'old_cont/old_tgt',
             'X-Symlink-Target-Account': 'a'}, None)
        cache = FakeCache({'sysmeta': {'versions-location': 'ver_cont'}})
        req = Request.blank(
            '/v1/a/c/symlink',
            headers={'X-Symlink-Target': 'new_cont/new_tgt'},
            environ={'REQUEST_METHOD': 'PUT', 'swift.cache': cache,
                     'CONTENT_LENGTH': '0',
                     'swift.trans_id': 'fake_trans_id'})
        status, headers, body = self.call_vw(req)
        self.assertEqual(status, '201 Created')
        # authorized twice now because versioned_writes now makes a check on
        # PUT
        self.assertEqual(len(self.authorized), 2)
        self.assertRequestEqual(req, self.authorized[0])
        self.assertEqual(['VW', 'VW', None], self.app.swift_sources)
        self.assertEqual({'fake_trans_id'}, set(self.app.txn_ids))
        calls = self.app.calls_with_headers
        method, path, req_headers = calls[2]
        self.assertEqual('PUT', method)
        self.assertEqual('/v1/a/c/symlink', path)
        self.assertEqual(
            'new_cont/new_tgt',
            req_headers['X-Object-Sysmeta-Symlink-Target'])

    def test_delete_latest_version_no_marker_success(self):
        self.app.register(
            'GET',
            '/v1/a/ver_cont?prefix=003sym/&marker=&reverse=on',
            swob.HTTPOk, {},
            '[{"hash": "y", '
            '"last_modified": "2014-11-21T14:23:02.206740", '
            '"bytes": 0, '
            '"name": "003sym/2", '
            '"content_type": "text/plain"}, '
            '{"hash": "x", '
            '"last_modified": "2014-11-21T14:14:27.409100", '
            '"bytes": 0, '
            '"name": "003sym/1", '
            '"content_type": "text/plain"}]')
        self.app.register(
            'GET', '/v1/a/ver_cont/003sym/2', swob.HTTPCreated,
            {'content-length': '0',
             'X-Object-Sysmeta-Symlink-Target': 'c/tgt'}, None)
        self.app.register(
            'PUT', '/v1/a/c/sym', swob.HTTPCreated,
            {'X-Symlink-Target': 'c/tgt', 'X-Symlink-Target-Account': 'a'},
            None)
        self.app.register(
            'DELETE', '/v1/a/ver_cont/003sym/2', swob.HTTPOk,
            {}, None)

        cache = FakeCache({'sysmeta': {'versions-location': 'ver_cont'}})
        req = Request.blank(
            '/v1/a/c/sym',
            headers={'X-If-Delete-At': 1},
            environ={'REQUEST_METHOD': 'DELETE', 'swift.cache': cache,
                     'CONTENT_LENGTH': '0', 'swift.trans_id': 'fake_trans_id'})
        status, headers, body = self.call_vw(req)
        self.assertEqual(status, '200 OK')
        self.assertEqual(len(self.authorized), 1)
        self.assertRequestEqual(req, self.authorized[0])
        self.assertEqual(4, self.app.call_count)
        self.assertEqual(['VW', 'VW', 'VW', 'VW'], self.app.swift_sources)
        self.assertEqual({'fake_trans_id'}, set(self.app.txn_ids))
        calls = self.app.calls_with_headers
        method, path, req_headers = calls[2]
        self.assertEqual('PUT', method)
        self.assertEqual('/v1/a/c/sym', path)
        self.assertEqual(
            'c/tgt',
            req_headers['X-Object-Sysmeta-Symlink-Target'])


class TestSymlinkContainerContext(TestSymlinkMiddlewareBase):

    def setUp(self):
        super(TestSymlinkContainerContext, self).setUp()
        self.context = symlink.SymlinkContainerContext(
            self.sym.app, self.sym.logger)

    def test_extract_symlink_path_json_simple_etag(self):
        obj_dict = {"bytes": 6,
                    "last_modified": "1",
                    "hash": "etag",
                    "name": "obj",
                    "content_type": "application/octet-stream"}
        obj_dict = self.context._extract_symlink_path_json(
            obj_dict, 'v1', 'AUTH_a')
        self.assertEqual(obj_dict['hash'], 'etag')
        self.assertNotIn('symlink_path', obj_dict)

    def test_extract_symlink_path_json_symlink_path(self):
        obj_dict = {"bytes": 6,
                    "last_modified": "1",
                    "hash": "etag; symlink_target=c/o; something_else=foo; "
                    "symlink_target_etag=tgt_etag; symlink_target_bytes=8",
                    "name": "obj",
                    "content_type": "application/octet-stream"}
        obj_dict = self.context._extract_symlink_path_json(
            obj_dict, 'v1', 'AUTH_a')
        self.assertEqual(obj_dict['hash'], 'etag; something_else=foo')
        self.assertEqual(obj_dict['symlink_path'], '/v1/AUTH_a/c/o')
        self.assertEqual(obj_dict['symlink_etag'], 'tgt_etag')
        self.assertEqual(obj_dict['symlink_bytes'], 8)

    def test_extract_symlink_path_json_symlink_path_and_account(self):
        obj_dict = {
            "bytes": 6,
            "last_modified": "1",
            "hash": "etag; symlink_target=c/o; symlink_target_account=AUTH_a2",
            "name": "obj",
            "content_type": "application/octet-stream"}
        obj_dict = self.context._extract_symlink_path_json(
            obj_dict, 'v1', 'AUTH_a')
        self.assertEqual(obj_dict['hash'], 'etag')
        self.assertEqual(obj_dict['symlink_path'], '/v1/AUTH_a2/c/o')

    def test_extract_symlink_path_json_extra_key(self):
        obj_dict = {"bytes": 6,
                    "last_modified": "1",
                    "hash": "etag; symlink_target=c/o; extra_key=value",
                    "name": "obj",
                    "content_type": "application/octet-stream"}
        obj_dict = self.context._extract_symlink_path_json(
            obj_dict, 'v1', 'AUTH_a')
        self.assertEqual(obj_dict['hash'], 'etag; extra_key=value')
        self.assertEqual(obj_dict['symlink_path'], '/v1/AUTH_a/c/o')

    def test_get_container_simple(self):
        self.app.register(
            'GET',
            '/v1/a/c',
            swob.HTTPOk, {},
            json.dumps(
                [{"hash": "etag; symlink_target=c/o;",
                  "last_modified": "2014-11-21T14:23:02.206740",
                  "bytes": 0,
                  "name": "sym_obj",
                  "content_type": "text/plain"},
                 {"hash": "etag2",
                  "last_modified": "2014-11-21T14:14:27.409100",
                  "bytes": 32,
                  "name": "normal_obj",
                  "content_type": "text/plain"}]))
        req = Request.blank(path='/v1/a/c')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')
        obj_list = json.loads(body)
        self.assertIn('symlink_path', obj_list[0])
        self.assertIn(obj_list[0]['symlink_path'], '/v1/a/c/o')
        self.assertNotIn('symlink_path', obj_list[1])

    def test_get_container_with_subdir(self):
        self.app.register(
            'GET',
            '/v1/a/c?delimiter=/',
            swob.HTTPOk, {},
            json.dumps([{"subdir": "photos/"}]))
        req = Request.blank(path='/v1/a/c?delimiter=/')
        status, headers, body = self.call_sym(req)
        self.assertEqual(status, '200 OK')
        obj_list = json.loads(body)
        self.assertEqual(len(obj_list), 1)
        self.assertEqual(obj_list[0]['subdir'], 'photos/')

    def test_get_container_error_cases(self):
        # No affect for error cases
        for error in (swob.HTTPNotFound, swob.HTTPUnauthorized,
                      swob.HTTPServiceUnavailable,
                      swob.HTTPInternalServerError):
            self.app.register('GET', '/v1/a/c', error, {}, '')
            req = Request.blank(path='/v1/a/c')
            status, headers, body = self.call_sym(req)
            self.assertEqual(status, error().status)

    def test_no_affect_for_account_request(self):
        with mock.patch.object(self.sym, 'app') as mock_app:
            mock_app.return_value = (b'ok',)
            req = Request.blank(path='/v1/a')
            status, headers, body = self.call_sym(req)
            self.assertEqual(body, b'ok')

    def test_get_container_simple_with_listing_format(self):
        self.app.register(
            'GET',
            '/v1/a/c?format=json',
            swob.HTTPOk, {},
            json.dumps(
                [{"hash": "etag; symlink_target=c/o;",
                  "last_modified": "2014-11-21T14:23:02.206740",
                  "bytes": 0,
                  "name": "sym_obj",
                  "content_type": "text/plain"},
                 {"hash": "etag2",
                  "last_modified": "2014-11-21T14:14:27.409100",
                  "bytes": 32,
                  "name": "normal_obj",
                  "content_type": "text/plain"}]))
        self.lf = listing_formats.filter_factory({})(self.sym)
        req = Request.blank(path='/v1/a/c?format=json')
        status, headers, body = self.call_app(req, app=self.lf)
        self.assertEqual(status, '200 OK')
        obj_list = json.loads(body)
        self.assertIn('symlink_path', obj_list[0])
        self.assertIn(obj_list[0]['symlink_path'], '/v1/a/c/o')
        self.assertNotIn('symlink_path', obj_list[1])

    def test_get_container_simple_with_listing_format_xml(self):
        self.app.register(
            'GET',
            '/v1/a/c?format=json',
            swob.HTTPOk, {'Content-Type': 'application/json'},
            json.dumps(
                [{"hash": "etag; symlink_target=c/o;",
                  "last_modified": "2014-11-21T14:23:02.206740",
                  "bytes": 0,
                  "name": "sym_obj",
                  "content_type": "text/plain"},
                 {"hash": "etag2",
                  "last_modified": "2014-11-21T14:14:27.409100",
                  "bytes": 32,
                  "name": "normal_obj",
                  "content_type": "text/plain"}]))
        self.lf = listing_formats.filter_factory({})(self.sym)
        req = Request.blank(path='/v1/a/c?format=xml')
        status, headers, body = self.call_app(req, app=self.lf)
        self.assertEqual(status, '200 OK')
        self.assertEqual(body.split(b'\n'), [
            b'<?xml version="1.0" encoding="UTF-8"?>',
            b'<container name="c"><object><name>sym_obj</name>'
            b'<hash>etag</hash><bytes>0</bytes>'
            b'<content_type>text/plain</content_type>'
            b'<last_modified>2014-11-21T14:23:02.206740</last_modified>'
            b'</object>'
            b'<object><name>normal_obj</name><hash>etag2</hash>'
            b'<bytes>32</bytes><content_type>text/plain</content_type>'
            b'<last_modified>2014-11-21T14:14:27.409100</last_modified>'
            b'</object></container>'])



# Copyright (c) 2010-2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import queue

import pickle  # nosec: B403
import errno
import os
import signal
import sys
import time
import uuid
from random import random, shuffle
from bisect import insort
from collections import deque

from eventlet import spawn, Timeout

from swift.common.bufferedhttp import http_connect
from swift.common.constraints import check_drive
from swift.common.exceptions import ConnectionTimeout
from swift.common.ring import Ring
from swift.common.utils import get_logger, renamer, write_pickle, \
    dump_recon_cache, config_true_value, RateLimitedIterator, split_path, \
    eventlet_monkey_patch, get_redirect_data, ContextPool, hash_path, \
    non_negative_float, config_positive_int_value, non_negative_int, \
    EventletRateLimiter, node_to_string, parse_options, load_recon_cache
from swift.common.daemon import Daemon, run_daemon
from swift.common.header_key_dict import HeaderKeyDict
from swift.common.storage_policy import split_policy_string, PolicyError
from swift.common.recon import RECON_OBJECT_FILE, DEFAULT_RECON_CACHE_PATH
from swift.obj.diskfile import get_tmp_dir, ASYNCDIR_BASE
from swift.common.http import is_success, HTTP_INTERNAL_SERVER_ERROR, \
    HTTP_MOVED_PERMANENTLY


class RateLimiterBucket(EventletRateLimiter):
    """
    Extends EventletRateLimiter to also maintain a deque of items that have
    been deferred due to rate-limiting, and to provide a comparator for sorting
    instanced by readiness.
    """
    def __init__(self, max_updates_per_second):
        super(RateLimiterBucket, self).__init__(max_updates_per_second,
                                                rate_buffer=0)
        self.deque = deque()

    def __len__(self):
        return len(self.deque)

    def __bool__(self):
        return bool(self.deque)

    def __lt__(self, other):
        # used to sort RateLimiterBuckets by readiness
        if isinstance(other, RateLimiterBucket):
            return self.running_time < other.running_time
        return self.running_time < other


class BucketizedUpdateSkippingLimiter(object):
    """
    Wrap an iterator to rate-limit updates on a per-bucket basis, where updates
    are mapped to buckets by hashing their destination path. If an update is
    rate-limited then it is placed on a deferral queue and may be sent later if
    the wrapped iterator is exhausted before the ``drain_until`` time is
    reached.

    The deferral queue has constrained size and once the queue is full updates
    are evicted using a first-in-first-out policy. This policy is used because
    updates on the queue may have been made obsolete by newer updates written
    to disk, and this is more likely for updates that have been on the queue
    longest.

    The iterator increments stats as follows:

    * The `deferrals` stat is incremented for each update that is
      rate-limited. Note that a individual update is rate-limited at most
      once.
    * The `skips` stat is incremented for each rate-limited update that is
      not eventually yielded. This includes updates that are evicted from the
      deferral queue and all updates that remain in the deferral queue when
      ``drain_until`` time is reached and the iterator terminates.
    * The `drains` stat is incremented for each rate-limited update that is
      eventually yielded.

    Consequently, when this iterator terminates, the sum of `skips` and
    `drains` is equal to the number of `deferrals`.

    :param update_iterable: an async_pending update iterable
    :param logger: a logger instance
    :param stats: a SweepStats instance
    :param num_buckets: number of buckets to divide container hashes into, the
                        more buckets total the less containers to a bucket
                        (once a busy container slows down a bucket the whole
                        bucket starts deferring)
    :param max_elements_per_group_per_second: tunable, when deferring kicks in
    :param max_deferred_elements: maximum number of deferred elements before
        skipping starts. Each bucket may defer updates, but once the total
        number of deferred updates summed across all buckets reaches this
        value then all buckets will skip subsequent updates.
    :param drain_until: time at which any remaining deferred elements must be
        skipped and the iterator stops. Once the wrapped iterator has been
        exhausted, this iterator will drain deferred elements from its buckets
        until either all buckets have drained or this time is reached.
    """

    def __init__(self, update_iterable, logger, stats, num_buckets=1000,
                 max_elements_per_group_per_second=50,
                 max_deferred_elements=0,
                 drain_until=0):
        self.iterator = iter(update_iterable)
        self.logger = logger
        self.stats = stats
        # if we want a smaller "blast radius" we could make this number bigger
        self.num_buckets = max(num_buckets, 1)
        self.max_deferred_elements = max_deferred_elements
        self.deferred_buckets = deque()
        self.drain_until = drain_until
        self.salt = str(uuid.uuid4())
        self.buckets = [RateLimiterBucket(max_elements_per_group_per_second)
                        for _ in range(self.num_buckets)]
        self.buckets_ordered_by_readiness = None

    def __iter__(self):
        return self

    def _bucket_key(self, update):
        acct, cont = split_update_path(update)
        return int(hash_path(acct, cont, self.salt), 16) % self.num_buckets

    def _get_time(self):
        return time.time()

    def __next__(self):
        # first iterate over the wrapped iterator...
        for update_ctx in self.iterator:
            bucket = self.buckets[self._bucket_key(update_ctx['update'])]
            now = self._get_time()
            if bucket.is_allowed(now=now):
                # no need to ratelimit, just return next update
                return update_ctx

            self.stats.deferrals += 1
            self.logger.increment("deferrals")
            if self.max_deferred_elements > 0:
                if len(self.deferred_buckets) >= self.max_deferred_elements:
                    # create space to defer this update by popping the least
                    # recent deferral from the least recently deferred bucket;
                    # updates read from disk recently are preferred over those
                    # read from disk less recently.
                    oldest_deferred_bucket = self.deferred_buckets.popleft()
                    oldest_deferred_bucket.deque.popleft()
                    self.stats.skips += 1
                    self.logger.increment("skips")
                # append the update to the bucket's queue and append the bucket
                # to the queue of deferred buckets
                # note: buckets may have multiple entries in deferred_buckets,
                # one for each deferred update in that particular bucket
                bucket.deque.append(update_ctx)
                self.deferred_buckets.append(bucket)
            else:
                self.stats.skips += 1
                self.logger.increment("skips")

        if self.buckets_ordered_by_readiness is None:
            # initialise a queue of those buckets with deferred elements;
            # buckets are queued in the chronological order in which they are
            # ready to serve an element
            self.buckets_ordered_by_readiness = queue.PriorityQueue()
            for bucket in self.buckets:
                if bucket:
                    self.buckets_ordered_by_readiness.put(bucket)

        # now drain the buckets...
        undrained_elements = []
        while not self.buckets_ordered_by_readiness.empty():
            now = self._get_time()
            bucket = self.buckets_ordered_by_readiness.get_nowait()
            if now < self.drain_until:
                # wait for next element to be ready
                bucket.wait(now=now)
                # drain the most recently deferred element
                item = bucket.deque.pop()
                if bucket:
                    # bucket has more deferred elements, re-insert in queue in
                    # correct chronological position
                    self.buckets_ordered_by_readiness.put(bucket)
                self.stats.drains += 1
                self.logger.increment("drains")
                return item
            else:
                # time to stop iterating: gather all un-drained elements
                undrained_elements.extend(bucket.deque)

        if undrained_elements:
            # report final batch of skipped elements
            self.stats.skips += len(undrained_elements)
            self.logger.update_stats("skips", len(undrained_elements))

        raise StopIteration()


class OldestAsyncPendingTracker:
    """
    Manages the tracking of the oldest async pending updates for each
    account-container pair using a sorted list for timestamps. Evicts the
    newest pairs when t max_entries is reached. Supports retrieving the N
    oldest async pending updates or calculating the age of the oldest pending
    update.
    """
    def __init__(
        self,
        max_entries,
    ):
        self.max_entries = max_entries
        self.sorted_entries = []
        self.ac_to_timestamp = {}

    def add_update(self, account, container, timestamp):
        """
        Add or update a timestamp for a given account and container.

        :param account: (str) The account name.
        :param container: (str) The container name.
        :param timestamp: (float) The timestamp to add or update.
        """
        # Ensure the timestamp is a float
        timestamp = float(timestamp)

        ac = (account, container)

        if ac in self.ac_to_timestamp:
            old_timestamp = self.ac_to_timestamp[ac]
            # Only replace the existing timestamp if the new one is older
            if timestamp < old_timestamp:
                # Remove the old (timestamp, ac) from the
                # sorted list
                self.sorted_entries.remove((old_timestamp, ac))
                # Insert the new (timestamp, ac) in the sorted order
                insort(self.sorted_entries, (timestamp, ac))
                # Update the ac_to_timestamp dictionary
                self.ac_to_timestamp[ac] = timestamp
        else:
            # Insert the new (timestamp, ac) in the sorted order
            insort(self.sorted_entries, (timestamp, ac))
            self.ac_to_timestamp[ac] = timestamp

        # Check size and evict the newest ac(s) if necessary
        if (len(self.ac_to_timestamp) > self.max_entries):
            # Pop the newest entry (largest timestamp)
            _, newest_ac = (self.sorted_entries.pop())
            del self.ac_to_timestamp[newest_ac]

    def get_n_oldest_timestamp_acs(self, n):
        oldest_entries = self.sorted_entries[:n]
        return {
            'oldest_count': len(oldest_entries),
            'oldest_entries': [
                {
                    'timestamp': entry[0],
                    'account': entry[1][0],
                    'container': entry[1][1],
                }
                for entry in oldest_entries
            ],
        }

    def get_oldest_timestamp(self):
        if self.sorted_entries:
            return self.sorted_entries[0][0]
        return None

    def get_oldest_timestamp_age(self):
        current_time = time.time()
        oldest_timestamp = self.get_oldest_timestamp()
        if oldest_timestamp is not None:
            return current_time - oldest_timestamp
        return None

    def reset(self):
        self.sorted_entries = []
        self.ac_to_timestamp = {}

    def get_memory_usage(self):
        return self._get_size(self)

    def _get_size(self, obj, seen=None):
        if seen is None:
            seen = set()

        obj_id = id(obj)
        if obj_id in seen:
            return 0
        seen.add(obj_id)

        size = sys.getsizeof(obj)

        if isinstance(obj, dict):
            size += sum(
                self._get_size(k, seen) + self._get_size(v, seen)
                for k, v in obj.items()
            )
        elif hasattr(obj, '__dict__'):
            size += self._get_size(obj.__dict__, seen)
        elif (
            hasattr(obj, '__iter__')
            and not isinstance(obj, (str, bytes, bytearray))
        ):
            size += sum(self._get_size(i, seen) for i in obj)

        return size


class SweepStats(object):
    """
    Stats bucket for an update sweep

    A measure of the rate at which updates are being rate-limited is::

        deferrals / (deferrals + successes + failures - drains)

    A measure of the rate at which updates are not being sent during a sweep
    is::

        skips / (skips + successes + failures)
    """
    def __init__(self, errors=0, failures=0, quarantines=0, successes=0,
                 unlinks=0, outdated_unlinks=0, redirects=0, skips=0,
                 deferrals=0, drains=0):
        self.errors = errors
        self.failures = failures
        self.quarantines = quarantines
        self.successes = successes
        self.unlinks = unlinks
        self.outdated_unlinks = outdated_unlinks
        self.redirects = redirects
        self.skips = skips
        self.deferrals = deferrals
        self.drains = drains

    def copy(self):
        return type(self)(self.errors, self.failures, self.quarantines,
                          self.successes, self.unlinks, self.outdated_unlinks,
                          self.redirects, self.skips, self.deferrals,
                          self.drains)

    def since(self, other):
        return type(self)(self.errors - other.errors,
                          self.failures - other.failures,
                          self.quarantines - other.quarantines,
                          self.successes - other.successes,
                          self.unlinks - other.unlinks,
                          self.outdated_unlinks - other.outdated_unlinks,
                          self.redirects - other.redirects,
                          self.skips - other.skips,
                          self.deferrals - other.deferrals,
                          self.drains - other.drains)

    def reset(self):
        self.errors = 0
        self.failures = 0
        self.quarantines = 0
        self.successes = 0
        self.unlinks = 0
        self.outdated_unlinks = 0
        self.redirects = 0
        self.skips = 0
        self.deferrals = 0
        self.drains = 0

    def __str__(self):
        keys = (
            (self.successes, 'successes'),
            (self.failures, 'failures'),
            (self.quarantines, 'quarantines'),
            (self.unlinks, 'unlinks'),
            (self.outdated_unlinks, 'outdated_unlinks'),
            (self.errors, 'errors'),
            (self.redirects, 'redirects'),
            (self.skips, 'skips'),
            (self.deferrals, 'deferrals'),
            (self.drains, 'drains'),
        )
        return ', '.join('%d %s' % pair for pair in keys)


def split_update_path(update):
    """
    Split the account and container parts out of the async update data.

    N.B. updates to shards set the container_path key while the account and
    container keys are always the root.
    """
    container_path = update.get('container_path')
    if container_path:
        acct, cont = split_path('/' + container_path, minsegs=2)
    else:
        acct, cont = update['account'], update['container']
    return acct, cont


class ObjectUpdater(Daemon):
    """Update object information in container listings."""

    def __init__(self, conf, logger=None):
        self.conf = conf
        self.logger = logger or get_logger(conf, log_route='object-updater')
        self.devices = conf.get('devices', '/srv/node')
        self.mount_check = config_true_value(conf.get('mount_check', 'true'))
        self.swift_dir = conf.get('swift_dir', '/etc/swift')
        self.interval = float(conf.get('interval', 300))
        self.container_ring = None
        self.concurrency = int(conf.get('concurrency', 8))
        self.updater_workers = config_positive_int_value(
            conf.get('updater_workers', 1))
        if 'slowdown' in conf:
            self.logger.warning(
                'The slowdown option is deprecated in favor of '
                'objects_per_second. This option may be ignored in a '
                'future release.')
            objects_per_second = 1 / (
                float(conf.get('slowdown', '0.01')) + 0.01)
        else:
            objects_per_second = 50
        self.objects_running_time = 0
        self.max_objects_per_second = \
            float(conf.get('objects_per_second',
                           objects_per_second))
        self.max_objects_per_container_per_second = non_negative_float(
            conf.get('max_objects_per_container_per_second', 0))
        self.per_container_ratelimit_buckets = config_positive_int_value(
            conf.get('per_container_ratelimit_buckets', 1000))
        self.node_timeout = float(conf.get('node_timeout', 10))
        self.conn_timeout = float(conf.get('conn_timeout', 0.5))
        self.report_interval = float(conf.get('report_interval', 300))
        self.recon_cache_path = conf.get('recon_cache_path',
                                         DEFAULT_RECON_CACHE_PATH)
        self.rcache = os.path.join(self.recon_cache_path, RECON_OBJECT_FILE)
        max_entries = config_positive_int_value(
            conf.get('async_tracker_max_entries', 100)
        )
        self.dump_count = config_positive_int_value(
            conf.get('async_tracker_dump_count', 5)
        )
        self.stats = SweepStats()
        self.oldest_async_pendings = OldestAsyncPendingTracker(max_entries)
        self.max_deferred_updates = non_negative_int(
            conf.get('max_deferred_updates', 10000))
        self.begin = time.time()

    def _listdir(self, path):
        try:
            return os.listdir(path)
        except OSError as e:
            self.stats.errors += 1
            self.logger.increment('errors')
            self.logger.error('ERROR: Unable to access %(path)s: '
                              '%(error)s',
                              {'path': path, 'error': e})
            return []

    def get_container_ring(self):
        """Get the container ring.  Load it, if it hasn't been yet."""
        if not self.container_ring:
            self.container_ring = Ring(self.swift_dir, ring_name='container')
        return self.container_ring

    def _process_device_in_child(self, dev_path, device):
        """Process a single device in a forked child process."""
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        os.environ.pop('NOTIFY_SOCKET', None)
        eventlet_monkey_patch()
        self.stats.reset()
        self.oldest_async_pendings.reset()
        forkbegin = time.time()
        self.object_sweep(dev_path)
        elapsed = time.time() - forkbegin
        self.logger.info(
            ('Object update sweep of %(device)s '
                'completed: %(elapsed).02fs, %(stats)s'),
            {'device': device, 'elapsed': elapsed,
                'stats': self.stats})
        self.dump_device_recon(device)

    def _process_devices(self, devices):
        """Process devices, handling both single and multi-threaded modes."""
        pids = []
        # read from container ring to ensure it's fresh
        self.get_container_ring().get_nodes('')
        for device in devices:
            try:
                dev_path = check_drive(self.devices, device,
                                       self.mount_check)
            except ValueError as err:
                # We don't count this as an error. The occasional
                # unmounted drive is part of normal cluster operations,
                # so a simple warning is sufficient.
                self.logger.warning('Skipping: %s', err)
                continue
            while len(pids) >= self.updater_workers:
                pids.remove(os.wait()[0])
            pid = os.fork()
            if pid:
                pids.append(pid)
            else:
                self._process_device_in_child(dev_path, device)
                sys.exit()

        while pids:
            pids.remove(os.wait()[0])

    def run_forever(self, *args, **kwargs):
        """Run the updater continuously."""
        time.sleep(random() * self.interval)
        while True:
            elapsed = self.run_once(*args, **kwargs)
            if elapsed < self.interval:
                time.sleep(self.interval - elapsed)

    def run_once(self, *args, **kwargs):
        """Run the updater once."""
        self.logger.info('Begin object update sweep of all devices')
        self.begin = time.time()
        devices = self._listdir(self.devices)
        self._process_devices(devices)
        now = time.time()
        elapsed = now - self.begin
        self.logger.info(
            ('Object update sweep of all devices completed: '
             '%(elapsed).02fs'),
            {'elapsed': elapsed})
        self.aggregate_and_dump_recon(devices, elapsed, now)
        return elapsed

    def _gather_recon_stats(self):
        """Gather stats for device recon dumps."""
        stats = {
            'failures_oldest_timestamp':
                self.oldest_async_pendings.get_oldest_timestamp(),
            'failures_oldest_timestamp_age':
                self.oldest_async_pendings.get_oldest_timestamp_age(),
            'failures_account_container_count':
                len(self.oldest_async_pendings.ac_to_timestamp),
            'failures_oldest_timestamp_account_containers':
                self.oldest_async_pendings.get_n_oldest_timestamp_acs(
                    self.dump_count),
            'tracker_memory_usage':
                self.oldest_async_pendings.get_memory_usage(),
        }
        return stats

    def dump_device_recon(self, device):
        """Dump recon stats for a single device."""
        disk_recon_stats = self._gather_recon_stats()
        dump_recon_cache(
            {'object_updater_per_device': {device: disk_recon_stats}},
            self.rcache,
            self.logger,
        )

    def aggregate_and_dump_recon(self, devices, elapsed, now):
        """
        Aggregate recon stats across devices and dump the result to the
        recon cache.
        """
        recon_cache = load_recon_cache(self.rcache)
        device_stats = recon_cache.get('object_updater_per_device', {})
        if not isinstance(device_stats, dict):
            raise TypeError('object_updater_per_device must be a dict')
        devices_to_remove = set(device_stats) - set(devices)
        device_stats = {dev: device_stats.get(dev) or {}
                        for dev in devices}

        aggregated_oldest_entries = []

        for stats in device_stats.values():
            container_data = stats.get(
                'failures_oldest_timestamp_account_containers', {})
            aggregated_oldest_entries.extend(container_data.get(
                'oldest_entries', []))
        aggregated_oldest_entries.sort(key=lambda x: x['timestamp'])
        aggregated_oldest_entries = aggregated_oldest_entries[:self.dump_count]
        aggregated_oldest_count = len(aggregated_oldest_entries)

        aggregated_stats = {
            'failures_account_container_count': max(
                list(
                    stats.get('failures_account_container_count', 0)
                    for stats in device_stats.values()
                )
                or [0],
            ),
            'tracker_memory_usage': (
                float(sum(
                    stats.get('tracker_memory_usage', 0)
                    for stats in device_stats.values()
                ))
                / float(len(device_stats))
            )
            * max(1, min(self.updater_workers, len(device_stats)))
            if device_stats
            else 0,
            'failures_oldest_timestamp': min(
                list(filter(lambda x: x is not None,
                            [stats.get('failures_oldest_timestamp')
                             for stats in device_stats.values()]
                            )) or [None],
            ),
            'failures_oldest_timestamp_age': max(
                list(filter(lambda x: x is not None,
                            [stats.get('failures_oldest_timestamp_age')
                             for stats in device_stats.values()]
                            )) or [None],
            ),
            'failures_oldest_timestamp_account_containers': {
                'oldest_count': aggregated_oldest_count,
                'oldest_entries': aggregated_oldest_entries,
            },
        }

        recon_dump = {
            'object_updater_sweep': elapsed,
            'object_updater_stats': aggregated_stats,
            'object_updater_last': now
        }
        if devices_to_remove:
            recon_dump['object_updater_per_device'] = {
                d: {} for d in devices_to_remove}
        dump_recon_cache(
            recon_dump,
            self.rcache,
            self.logger,
        )

    def _load_update(self, device, update_path):
        try:
            return pickle.load(open(update_path, 'rb'))  # nosec: B301
        except Exception as e:
            if getattr(e, 'errno', None) == errno.ENOENT:
                return
            self.logger.exception(
                'ERROR Pickle problem, quarantining %s', update_path)
            self.stats.quarantines += 1
            self.logger.increment('quarantines')
            target_path = os.path.join(device, 'quarantined', 'objects',
                                       os.path.basename(update_path))
            renamer(update_path, target_path, fsync=False)
            try:
                # If this was the last async_pending in the directory,
                # then this will succeed. Otherwise, it'll fail, and
                # that's okay.
                os.rmdir(os.path.dirname(update_path))
            except OSError:
                pass
            return

    def _iter_async_pendings(self, device):
        """
        Locate and yield an update context for all the async pending files on
        the device. Each update context contains details of the async pending
        file location, its timestamp and the un-pickled update data.

        Async pending files that fail to load will be quarantined.

        Only the most recent update for the same object is yielded; older
        (stale) async pending files are unlinked as they are located.

        The iterator tries to clean up empty directories as it goes.
        """
        # loop through async pending dirs for all policies
        for asyncdir in self._listdir(device):
            # we only care about directories
            async_pending = os.path.join(device, asyncdir)
            if not asyncdir.startswith(ASYNCDIR_BASE):
                # skip stuff like "accounts", "containers", etc.
                continue
            if not os.path.isdir(async_pending):
                continue
            try:
                base, policy = split_policy_string(asyncdir)
            except PolicyError as e:
                # This isn't an error, but a misconfiguration. Logging a
                # warning should be sufficient.
                self.logger.warning('Directory %(directory)r does not map '
                                    'to a valid policy (%(error)s)', {
                                        'directory': asyncdir, 'error': e})
                continue
            prefix_dirs = self._listdir(async_pending)
            shuffle(prefix_dirs)
            for prefix in prefix_dirs:
                prefix_path = os.path.join(async_pending, prefix)
                if not os.path.isdir(prefix_path):
                    continue
                last_obj_hash = None
                for update_file in sorted(self._listdir(prefix_path),
                                          reverse=True):
                    update_path = os.path.join(prefix_path, update_file)
                    if not os.path.isfile(update_path):
                        continue
                    try:
                        obj_hash, timestamp = update_file.split('-')
                    except ValueError:
                        self.stats.errors += 1
                        self.logger.increment('errors')
                        self.logger.error(
                            'ERROR async pending file with unexpected '
                            'name %s', update_path)
                        continue
                    # Async pendings are stored on disk like this:
                    #
                    # <device>/async_pending/<suffix>/<obj_hash>-<timestamp>
                    #
                    # If there are multiple updates for a given object,
                    # they'll look like this:
                    #
                    # <device>/async_pending/<obj_suffix>/<obj_hash>-<timestamp1>
                    # <device>/async_pending/<obj_suffix>/<obj_hash>-<timestamp2>
                    # <device>/async_pending/<obj_suffix>/<obj_hash>-<timestamp3>
                    #
                    # Async updates also have the property that newer
                    # updates contain all the information in older updates.
                    # Since we sorted the directory listing in reverse
                    # order, we'll see timestamp3 first, yield it, and then
                    # unlink timestamp2 and timestamp1 since we know they
                    # are obsolete.
                    #
                    # This way, our caller only gets useful async_pendings.
                    if obj_hash == last_obj_hash:
                        self.stats.outdated_unlinks += 1
                        self.logger.increment('outdated_unlinks')
                        try:
                            os.unlink(update_path)
                        except OSError as e:
                            if e.errno != errno.ENOENT:
                                raise
                    else:
                        last_obj_hash = obj_hash
                        update = self._load_update(device, update_path)
                        if update is not None:
                            yield {'device': device,
                                   'policy': policy,
                                   'update_path': update_path,
                                   'obj_hash': obj_hash,
                                   'timestamp': timestamp,
                                   'update': update}

    def object_sweep(self, device):
        """
        If there are async pendings on the device, walk each one and update.

        :param device: path to device
        """
        start_time = time.time()
        last_status_update = start_time
        start_stats = self.stats.copy()
        my_pid = os.getpid()
        self.logger.info("Object update sweep starting on %s (pid: %d)",
                         device, my_pid)

        ap_iter = RateLimitedIterator(
            self._iter_async_pendings(device),
            elements_per_second=self.max_objects_per_second)
        ap_iter = BucketizedUpdateSkippingLimiter(
            ap_iter, self.logger, self.stats,
            self.per_container_ratelimit_buckets,
            self.max_objects_per_container_per_second,
            max_deferred_elements=self.max_deferred_updates,
            drain_until=self.begin + self.interval)
        with ContextPool(self.concurrency) as pool:
            for update_ctx in ap_iter:
                pool.spawn(self.process_object_update, **update_ctx)
                now = time.time()
                if now - last_status_update >= self.report_interval:
                    this_sweep = self.stats.since(start_stats)
                    self.logger.info(
                        ('Object update sweep progress on %(device)s: '
                         '%(elapsed).02fs, %(stats)s (pid: %(pid)d)'),
                        {'device': device,
                         'elapsed': now - start_time,
                         'pid': my_pid,
                         'stats': this_sweep})
                    last_status_update = now
            pool.waitall()

        self.logger.timing_since('timing', start_time)
        sweep_totals = self.stats.since(start_stats)
        self.logger.info(
            ('Object update sweep completed on %(device)s '
             'in %(elapsed).02fs seconds:, '
             '%(successes)d successes, %(failures)d failures, '
             '%(quarantines)d quarantines, '
             '%(unlinks)d unlinks, '
             '%(outdated_unlinks)d outdated_unlinks, '
             '%(errors)d errors, '
             '%(redirects)d redirects, '
             '%(skips)d skips, '
             '%(deferrals)d deferrals, '
             '%(drains)d drains '
             '(pid: %(pid)d)'),
            {'device': device,
             'elapsed': time.time() - start_time,
             'pid': my_pid,
             'successes': sweep_totals.successes,
             'failures': sweep_totals.failures,
             'quarantines': sweep_totals.quarantines,
             'unlinks': sweep_totals.unlinks,
             'outdated_unlinks': sweep_totals.outdated_unlinks,
             'errors': sweep_totals.errors,
             'redirects': sweep_totals.redirects,
             'skips': sweep_totals.skips,
             'deferrals': sweep_totals.deferrals,
             'drains': sweep_totals.drains
             })

    def process_object_update(self, update_path, device, policy, update,
                              **kwargs):
        """
        Process the object information to be updated and update.

        :param update_path: path to pickled object update file
        :param device: path to device
        :param policy: storage policy of object update
        :param update: the un-pickled update data
        :param kwargs: un-used keys from update_ctx
        """

        def do_update():
            successes = update.get('successes', [])
            headers_out = HeaderKeyDict(update['headers'].copy())
            headers_out['user-agent'] = 'object-updater %s' % os.getpid()
            headers_out.setdefault('X-Backend-Storage-Policy-Index',
                                   str(int(policy)))
            headers_out.setdefault('X-Backend-Accept-Redirect', 'true')
            headers_out.setdefault('X-Backend-Accept-Quoted-Location', 'true')
            acct, cont = split_update_path(update)
            part, nodes = self.get_container_ring().get_nodes(acct, cont)
            path = '/%s/%s/%s' % (acct, cont, update['obj'])
            events = [spawn(self.object_update,
                            node, part, update['op'], path, headers_out)
                      for node in nodes if node['id'] not in successes]
            success = True
            new_successes = rewrite_pickle = False
            redirect = None
            redirects = set()
            for event in events:
                event_success, node_id, redirect = event.wait()
                if event_success is True:
                    successes.append(node_id)
                    new_successes = True
                else:
                    success = False
                if redirect:
                    redirects.add(redirect)

            if success:
                self.stats.successes += 1
                self.logger.increment('successes')
                self.logger.debug('Update sent for %(path)s %(update_path)s',
                                  {'path': path, 'update_path': update_path})
                self.stats.unlinks += 1
                self.logger.increment('unlinks')
                os.unlink(update_path)
                try:
                    # If this was the last async_pending in the directory,
                    # then this will succeed. Otherwise, it'll fail, and
                    # that's okay.
                    os.rmdir(os.path.dirname(update_path))
                except OSError:
                    pass
            elif redirects:
                # erase any previous successes
                update.pop('successes', None)
                redirect = max(redirects, key=lambda x: x[-1])[0]
                redirect_history = update.setdefault('redirect_history', [])
                if redirect in redirect_history:
                    # force next update to be sent to root, reset history
                    update['container_path'] = None
                    update['redirect_history'] = []
                else:
                    update['container_path'] = redirect
                    redirect_history.append(redirect)
                self.stats.redirects += 1
                self.logger.increment("redirects")
                self.logger.debug(
                    'Update redirected for %(path)s %(update_path)s to '
                    '%(shard)s',
                    {'path': path, 'update_path': update_path,
                     'shard': update['container_path']})
                rewrite_pickle = True
            else:
                self.stats.failures += 1
                self.logger.increment('failures')
                self.logger.debug('Update failed for %(path)s %(update_path)s',
                                  {'path': path, 'update_path': update_path})
                self.oldest_async_pendings.add_update(
                    acct, cont, kwargs['timestamp']
                )
                if new_successes:
                    update['successes'] = successes
                    rewrite_pickle = True

            return rewrite_pickle, redirect

        rewrite_pickle, redirect = do_update()
        if redirect:
            # make one immediate retry to the redirect location
            rewrite_pickle, redirect = do_update()
        if rewrite_pickle:
            write_pickle(update, update_path, os.path.join(
                device, get_tmp_dir(policy)))

    def object_update(self, node, part, op, path, headers_out):
        """
        Perform the object update to the container

        :param node: node dictionary from the container ring
        :param part: partition that holds the container
        :param op: operation performed (ex: 'PUT' or 'DELETE')
        :param path: /<acct>/<cont>/<obj> path being updated
        :param headers_out: headers to send with the update
        :return: a tuple of (``success``, ``node_id``, ``redirect``)
            where ``success`` is True if the update succeeded, ``node_id`` is
            the_id of the node updated and ``redirect`` is either None or a
            tuple of (a path, a timestamp string).
        """
        redirect = None
        start = time.time()
        # Assume an error until we hear otherwise
        status = 500
        try:
            with ConnectionTimeout(self.conn_timeout):
                conn = http_connect(
                    node['replication_ip'], node['replication_port'],
                    node['device'], part, op, path, headers_out)
            with Timeout(self.node_timeout):
                resp = conn.getresponse()
                resp.read()
            status = resp.status

            if status == HTTP_MOVED_PERMANENTLY:
                try:
                    redirect = get_redirect_data(resp)
                except ValueError as err:
                    self.logger.error(
                        'Container update failed for %r; problem with '
                        'redirect location: %s' % (path, err))

            success = is_success(status)
            if not success:
                self.logger.debug(
                    'Error code %(status)d is returned from remote '
                    'server %(node)s',
                    {'status': resp.status,
                     'node': node_to_string(node, replication=True)})
            return success, node['id'], redirect
        except Exception:
            self.logger.exception('ERROR with remote server %s',
                                  node_to_string(node, replication=True))
        except Timeout as exc:
            action = 'connecting to'
            if not isinstance(exc, ConnectionTimeout):
                # i.e., we definitely made the request but gave up
                # waiting for the response
                status = 499
                action = 'waiting on'
            self.logger.info(
                'Timeout %s remote server %s: %s',
                action, node_to_string(node, replication=True), exc)
        finally:
            elapsed = time.time() - start
            self.logger.timing('updater.timing.status.%s' % status,
                               elapsed * 1000)
        return HTTP_INTERNAL_SERVER_ERROR, node['id'], redirect


def main():
    conf_file, options = parse_options(once=True)
    run_daemon(ObjectUpdater, conf_file, **options)


if __name__ == '__main__':
    main()



# Copyright (c) 2022 NVIDIA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import time

from swift.common.request_helpers import split_and_validate_path
from swift.common.swob import Request, HTTPTooManyBackendRequests, \
    HTTPException
from swift.common.utils import get_logger, non_negative_float, \
    EventletRateLimiter, readconf

RATE_LIMITED_METHODS = ('GET', 'HEAD', 'PUT', 'POST', 'DELETE', 'UPDATE',
                        'REPLICATE')
BACKEND_RATELIMIT_CONFIG_SECTION = 'backend_ratelimit'
DEFAULT_BACKEND_RATELIMIT_CONF_FILE = 'backend-ratelimit.conf'
DEFAULT_CONFIG_RELOAD_INTERVAL = 60.0
DEFAULT_REQUESTS_PER_DEVICE_PER_SECOND = 0.0
DEFAULT_REQUESTS_PER_DEVICE_RATE_BUFFER = 1.0


class BackendRateLimitMiddleware(object):
    """
    Backend rate-limiting middleware.

    Rate-limits requests to backend storage node devices. Each (device, request
    method) combination is independently rate-limited. All requests with a
    'GET', 'HEAD', 'PUT', 'POST', 'DELETE', 'UPDATE' or 'REPLICATE' method are
    rate limited on a per-device basis by both a method-specific rate and an
    overall device rate limit.

    If a request would cause the rate-limit to be exceeded for the method
    and/or device then a response with a 529 status code is returned.
    """
    def __init__(self, app, filter_conf, logger=None):
        self.app = app
        self.filter_conf = filter_conf
        self.logger = logger or get_logger(self.filter_conf,
                                           log_route='backend_ratelimit')
        self.requests_per_device_rate_buffer = \
            DEFAULT_REQUESTS_PER_DEVICE_RATE_BUFFER
        # map (device, method) -> rate
        self.requests_per_device_per_second = {}
        # map (device, method) -> RateLimiter, populated on-demand
        self.rate_limiters = {}

        # some config options are *only* read from filter conf at startup...
        default_conf_path = os.path.join(
            self.filter_conf.get('swift_dir', '/etc/swift'),
            DEFAULT_BACKEND_RATELIMIT_CONF_FILE)
        try:
            self.conf_path = self.filter_conf['backend_ratelimit_conf_path']
            self.is_config_file_expected = True
        except KeyError:
            self.conf_path = default_conf_path
            self.is_config_file_expected = False
        self.config_reload_interval = non_negative_float(
            filter_conf.get('config_reload_interval',
                            DEFAULT_CONFIG_RELOAD_INTERVAL))

        # other conf options are read from filter section at startup but may
        # also be overridden by options in a separate config file...
        self._last_config_reload_attempt = time.time()
        self._apply_config(self.filter_conf)
        self._load_config_file()

    def _refresh_ratelimiters(self):
        # note: if we ever wanted to prune the ratelimiters (in case devices
        # have been removed) we could inspect each ratelimiter's running_time
        # and remove those with very old running_time
        for (dev, method), rl in self.rate_limiters.items():
            rl.set_max_rate(self.requests_per_device_per_second[method])
            rl.set_rate_buffer(self.requests_per_device_rate_buffer)

    def _apply_config(self, conf):
        modified = False
        reqs_per_device_rate_buffer = non_negative_float(
            conf.get('requests_per_device_rate_buffer',
                     DEFAULT_REQUESTS_PER_DEVICE_RATE_BUFFER))

        # note: 'None' key holds the aggregate per-device limit for all methods
        reqs_per_device_per_second = {None: non_negative_float(
            conf.get('requests_per_device_per_second', 0.0))}
        for method in RATE_LIMITED_METHODS:
            val = non_negative_float(
                conf.get('%s_requests_per_device_per_second'
                         % method.lower(), 0.0))
            reqs_per_device_per_second[method] = val

        if reqs_per_device_rate_buffer != self.requests_per_device_rate_buffer:
            self.requests_per_device_rate_buffer = reqs_per_device_rate_buffer
            modified = True
        if reqs_per_device_per_second != self.requests_per_device_per_second:
            self.requests_per_device_per_second = reqs_per_device_per_second
            self.is_any_rate_limit_configured = any(
                self.requests_per_device_per_second.values())
            modified = True
        if modified:
            self._refresh_ratelimiters()
        return modified

    def _load_config_file(self):
        # If conf file can be read then apply its options to the filter conf
        # options, discarding *all* options previously loaded from the conf
        # file i.e. options deleted from the conf file will revert to the
        # filter conf value or default value. If the conf file cannot be read
        # or is invalid, then the current config is left unchanged.
        try:
            new_conf = dict(self.filter_conf)  # filter_conf not current conf
            new_conf.update(
                readconf(self.conf_path, BACKEND_RATELIMIT_CONFIG_SECTION))
            modified = self._apply_config(new_conf)
            if modified:
                self.logger.info('Loaded config file %s, config changed',
                                 self.conf_path)
            elif not self.is_config_file_expected:
                self.logger.info('Loaded new config file %s, config unchanged',
                                 self.conf_path)
            else:
                self.logger.debug(
                    'Loaded existing config file %s, config unchanged',
                    self.conf_path)
            self.is_config_file_expected = True
        except IOError as err:
            if self.is_config_file_expected:
                self.logger.warning(
                    'Failed to load config file, config unchanged: %s', err)
            self.is_config_file_expected = False
        except ValueError as err:
            # ...but if it exists it should be valid
            self.logger.warning('Invalid config file %s, config unchanged: %s',
                                self.conf_path, err)

    def _maybe_reload_config(self):
        if self.config_reload_interval:
            now = time.time()
            if (now - self._last_config_reload_attempt
                    >= self.config_reload_interval):
                try:
                    self._load_config_file()
                except Exception:  # noqa
                    self.logger.exception('Error reloading config file')
                finally:
                    # always reset last loaded time to avoid re-try storm
                    self._last_config_reload_attempt = now

    def _get_ratelimiter(self, device, method=None):
        """
        Get a rate limiter for the (device, method) combination. If a rate
        limiter does not yet exist for the given (device, method) combination
        then it is created and added to the map of rate limiters.

        :param: the device.
        :method: the request method; if None then the aggregate rate limiter
            for all requests to the device is returned.
        :returns: an instance of ``EventletRateLimiter``.
        """
        try:
            rl = self.rate_limiters[(device, method)]
        except KeyError:
            rl = EventletRateLimiter(
                max_rate=self.requests_per_device_per_second[method],
                rate_buffer=self.requests_per_device_rate_buffer,
                running_time=time.time(),
                burst_after_idle=True)
            self.rate_limiters[(device, method)] = rl
        return rl

    def _is_allowed(self, device, method):
        """
        Evaluate backend rate-limiting policies for the incoming request.

        A request is allowed when neither the per-(device, method) rate-limit
        nor the per-device rate-limit has been reached.

        Note: a request will be disallowed if the aggregate per-device
        rate-limit has been reached, even if the per-(device, method)
        rate-limit has not been reached for the request's method.

        :param: the device.
        :method: the request method.
        :returns: boolean, is_allowed.
        """
        return (self._get_ratelimiter(device, None).is_allowed()
                and self._get_ratelimiter(device, method).is_allowed())

    def __call__(self, env, start_response):
        """
        WSGI entry point.

        :param env: WSGI environment dictionary
        :param start_response: WSGI callable
        """
        self._maybe_reload_config()
        req = Request(env)
        handler = self.app
        if (self.is_any_rate_limit_configured
                and req.method in RATE_LIMITED_METHODS):
            try:
                device, partition, _ = split_and_validate_path(req, 1, 3, True)
                int(partition)  # check it's a valid partition
            except (ValueError, HTTPException):
                # request may not have device/partition e.g. a healthcheck req
                pass
            else:
                if not self._is_allowed(device, req.method):
                    self.logger.increment('backend.ratelimit')
                    handler = HTTPTooManyBackendRequests()
        return handler(env, start_response)


def filter_factory(global_conf, **local_conf):
    conf = global_conf.copy()
    conf.update(local_conf)

    def backend_ratelimit_filter(app):
        return BackendRateLimitMiddleware(app, conf)

    return backend_ratelimit_filter



# Copyright (c) 2023 NVIDIA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import binascii
import string
import json
from unittest import mock

from swift.common import swob, utils
from swift.common.request_helpers import get_reserved_name
from swift.common.middleware import symlink
from swift.common.middleware.versioned_writes import object_versioning as ov

from test.unit import make_timestamp_iter
from test.unit.common.middleware.test_slo import slo, md5hex
from test.unit.common.middleware.s3api import (
    S3ApiTestCase, S3ApiTestCaseAcl, _gen_test_headers)


def _prepare_mpu(swift, ts_iter, upload_id, num_segments,
                 segment_bucket='bucket+segments', segment_key='mpu'):
    manifest = []
    for i, letter in enumerate(string.ascii_lowercase):
        if len(manifest) >= num_segments:
            break
        size = (i + 1) * 5
        body = letter * size
        etag = md5hex(body)
        path = '/%s/%s/%s/%s' % (segment_bucket, segment_key, upload_id, i + 1)
        swift.register('GET', '/v1/AUTH_test' + path, swob.HTTPOk, {
            'Content-Length': len(body),
            'Etag': etag,
        }, body)
        manifest.append({
            "name": path,
            "bytes": size,
            "hash": etag,
            "content_type": "application/octet-stream",
            "last_modified": next(ts_iter).isoformat,
        })
    slo_etag = md5hex(''.join(s['hash'] for s in manifest))
    s3_hash = md5hex(binascii.a2b_hex(''.join(
        s['hash'] for s in manifest)))
    s3_etag = "%s-%s" % (s3_hash, len(manifest))
    manifest_json = json.dumps(manifest)
    json_md5 = md5hex(manifest_json)
    manifest_headers = {
        'Content-Length': str(len(manifest_json)),
        'X-Static-Large-Object': 'true',
        'Etag': json_md5,
        'Content-Type': 'application/octet-stream',
        'X-Object-Sysmeta-Slo-Etag': slo_etag,
        'X-Object-Sysmeta-Slo-Size': str(sum(
            s['bytes'] for s in manifest)),
        'X-Object-Sysmeta-S3Api-Etag': s3_etag,
        'X-Object-Sysmeta-S3Api-Upload-Id': upload_id,
        'X-Object-Sysmeta-Container-Update-Override-Etag':
        '%s; s3_etag=%s; slo_etag=%s' % (json_md5, s3_etag, slo_etag),
    }
    return manifest_headers, manifest_json


class TestMpuGETorHEAD(S3ApiTestCase):

    def _wrap_app(self, app):
        self.slo = slo.filter_factory({'rate_limit_under_size': '0'})(app)
        return super(TestMpuGETorHEAD, self)._wrap_app(self.slo)

    def setUp(self):
        # this will call our _wrap_app
        super(TestMpuGETorHEAD, self).setUp()
        self.ts = make_timestamp_iter()
        manifest_headers, manifest_json = _prepare_mpu(
            self.swift, self.ts, 'X', 3)
        self.s3_etag = manifest_headers['X-Object-Sysmeta-S3Api-Etag']
        self.swift.register(
            'GET', '/v1/AUTH_test/bucket/mpu',
            swob.HTTPOk, manifest_headers, manifest_json.encode('ascii'))
        self.s3_acl = False

    def test_mpu_GET(self):
        req = swob.Request.blank('/bucket/mpu', headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status.split()[0], '200')
        self.assertEqual(body, b'aaaaabbbbbbbbbbccccccccccccccc')
        expected_calls = [
            ('GET', '/v1/AUTH_test/bucket/mpu'),
            ('GET', '/v1/AUTH_test/bucket+segments/mpu/X/1'
             '?multipart-manifest=get'),
            ('GET', '/v1/AUTH_test/bucket+segments/mpu/X/2'
             '?multipart-manifest=get'),
            ('GET', '/v1/AUTH_test/bucket+segments/mpu/X/3'
             '?multipart-manifest=get'),
        ]
        if self.s3_acl:
            # pre-flight object ACL check
            expected_calls.insert(0, ('HEAD', '/v1/AUTH_test/bucket/mpu'))
        self.assertEqual(self.swift.calls, expected_calls)
        self.assertEqual(headers['Content-Length'], '30')
        self.assertEqual(headers['Etag'], '"%s"' % self.s3_etag)
        self.assertNotIn('X-Amz-Mp-Parts-Count', headers)

    def test_mpu_GET_part_num(self):
        req = swob.Request.blank('/bucket/mpu', params={
            'partNumber': '2',
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status.split()[0], '206')
        self.assertEqual(body, b'bbbbbbbbbb')
        expected_calls = [
            ('GET', '/v1/AUTH_test/bucket/mpu?part-number=2'),
            ('GET', '/v1/AUTH_test/bucket+segments/mpu/X/2'
             '?multipart-manifest=get'),
        ]
        if self.s3_acl:
            expected_calls.insert(0, ('HEAD', '/v1/AUTH_test/bucket/mpu'))
        self.assertEqual(self.swift.calls, expected_calls)
        self.assertEqual(headers['Content-Length'], '10')
        self.assertEqual(headers['Content-Range'], 'bytes 5-14/30')
        self.assertEqual(headers['Etag'], '"%s"' % self.s3_etag)
        self.assertEqual(headers['X-Amz-Mp-Parts-Count'], '3')

    def test_mpu_GET_invalid_part_num(self):
        req = swob.Request.blank('/bucket/mpu', params={
            'partNumber': 'foo',
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status.split()[0], '400')
        self.assertEqual(self._get_error_code(body), 'InvalidArgument')
        self.assertEqual(self.swift.calls, [])

    def test_mpu_GET_zero_part_num(self):
        req = swob.Request.blank('/bucket/mpu', params={
            'partNumber': '0',
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status.split()[0], '400')
        self.assertEqual(self._get_error_code(body), 'InvalidArgument')
        self.assertEqual(self.swift.calls, [])

    def _do_test_mpu_GET_out_of_range_part_num(self, part_number):
        self.swift.clear_calls()
        req = swob.Request.blank('/bucket/mpu', params={
            'partNumber': str(part_number),
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status.split()[0], '416')
        self.assertEqual(self._get_error_code(body), 'InvalidPartNumber')
        expected_calls = [
            # s3api.controller.obj doesn't know yet if it's SLO, we delegate
            # param validation
            ('GET', '/v1/AUTH_test/bucket/mpu?part-number=%s' % part_number),
        ]
        if self.s3_acl:
            expected_calls.insert(0, ('HEAD', '/v1/AUTH_test/bucket/mpu'))
        self.assertEqual(self.swift.calls, expected_calls)

    def test_mpu_GET_out_of_range_part_num(self):
        self._do_test_mpu_GET_out_of_range_part_num(4)
        self._do_test_mpu_GET_out_of_range_part_num(10000)

    def test_existing_part_number_greater_than_max_parts_allowed(self):
        part_number = 3
        max_parts = 2
        req = swob.Request.blank('/bucket/mpu', params={
            'partNumber': str(part_number),
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        bad_req = swob.Request.blank('/bucket/mpu', params={
            'partNumber': str(part_number + 1),
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        with mock.patch.object(self.s3api.conf,
                               'max_upload_part_num', max_parts):
            # num_parts >= part number > max parts
            status, headers, body = self.call_s3api(req)
            self.assertEqual(status.split()[0], '206')
            # part number > num parts > max parts
            status, headers, body = self.call_s3api(bad_req)
            self.assertEqual(status.split()[0], '400')
            self.assertIn('must be an integer between 1 and 3, inclusive',
                          self._get_error_message(body))

        max_parts = part_number + 1
        with mock.patch.object(self.s3api.conf,
                               'max_upload_part_num', max_parts):
            # max_parts > num_parts >= part number
            status, headers, body = self.call_s3api(req)
            self.assertEqual(status.split()[0], '206')
            # max_parts >= part number > num parts
            status, headers, body = self.call_s3api(bad_req)
            self.assertEqual(status.split()[0], '416')
            self.assertIn('The requested partnumber is not satisfiable',
                          self._get_error_message(body))
            # part number > max_parts > num parts
            bad_req.params = {'partNumber': str(max_parts + 1)}
            status, headers, body = self.call_s3api(bad_req)
            self.assertEqual(status.split()[0], '400')
            self.assertIn('must be an integer between 1 and 4, inclusive',
                          self._get_error_message(body))

    def test_mpu_GET_huge_part_num(self):
        req = swob.Request.blank('/bucket/mpu', params={
            'partNumber': '10001',
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status.split()[0], '400')
        self.assertEqual(self._get_error_code(body), 'InvalidArgument')
        expected_calls = [
            # XXX is this value configurable?  do we need the SLO request?
            ('GET', '/v1/AUTH_test/bucket/mpu?part-number=10001'),
        ]
        if self.s3_acl:
            expected_calls.insert(0, ('HEAD', '/v1/AUTH_test/bucket/mpu'))
        self.assertEqual(self.swift.calls, expected_calls)

    def test_mpu_HEAD_part_num(self):
        req = swob.Request.blank('/bucket/mpu', params={
            'partNumber': '1',
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        }, method='HEAD')
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status.split()[0], '206')
        self.assertEqual(body, b'')
        self.assertEqual(self.swift.calls, [
            ('HEAD', '/v1/AUTH_test/bucket/mpu?part-number=1'),
            ('GET', '/v1/AUTH_test/bucket/mpu?part-number=1'),
        ])
        self.assertEqual(headers['Content-Length'], '5')
        self.assertEqual(headers['Content-Range'], 'bytes 0-4/30')
        self.assertEqual(headers['Etag'], '"%s"' % self.s3_etag)
        self.assertEqual(headers['X-Amz-Mp-Parts-Count'], '3')

    def test_mpu_HEAD_invalid_part_num(self):
        req = swob.Request.blank('/bucket/mpu', method='HEAD', params={
            'partNumber': 'foo',
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, _ = self.call_s3api(req)
        self.assertEqual(status.split()[0], '400')
        self.assertEqual(self.swift.calls, [])

    def test_mpu_HEAD_zero_part_num(self):
        req = swob.Request.blank('/bucket/mpu', method='HEAD', params={
            'partNumber': '0',
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, _ = self.call_s3api(req)
        self.assertEqual(status.split()[0], '400')
        self.assertEqual(self.swift.calls, [])

    def _do_test_mpu_HEAD_out_of_range_part_num(self, part_number):
        self.swift.clear_calls()
        req = swob.Request.blank('/bucket/mpu', method='HEAD', params={
            'partNumber': str(part_number),
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, _ = self.call_s3api(req)
        self.assertEqual(status.split()[0], '416')
        self.assertEqual(self.swift.calls, [
            ('HEAD', '/v1/AUTH_test/bucket/mpu?part-number=%s' % part_number),
            # SLO has to refetch to *see* if it's out-of-bounds
            ('GET', '/v1/AUTH_test/bucket/mpu?part-number=%s' % part_number),
        ])

    def test_mpu_HEAD_out_of_range_part_num(self):
        self._do_test_mpu_HEAD_out_of_range_part_num(4)
        self._do_test_mpu_HEAD_out_of_range_part_num(10000)

    def test_mpu_HEAD_huge_part_num(self):
        req = swob.Request.blank('/bucket/mpu', method='HEAD', params={
            'partNumber': '10001',
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status.split()[0], '400')
        self.assertEqual(self.swift.calls, [
            ('HEAD', '/v1/AUTH_test/bucket/mpu?part-number=10001'),
            # XXX were two requests worth it to 400?
            # how big can you configure SLO?
            # do such manifests *exist*?
            ('GET', '/v1/AUTH_test/bucket/mpu?part-number=10001'),
        ])


class TestMpuGETorHEADAcl(TestMpuGETorHEAD, S3ApiTestCaseAcl):

    def setUp(self):
        super(TestMpuGETorHEADAcl, self).setUp()
        object_headers = _gen_test_headers(
            self.default_owner, self.grants, 'object')
        self.swift.update_sticky_response_headers(
            '/v1/AUTH_test/bucket/mpu', object_headers)
        # this is used to flag insertion of expected HEAD pre-flight request of
        # object ACLs
        self.s3_acl = True


class TestVersionedMpuGETorHEAD(S3ApiTestCase):

    def _wrap_app(self, app):
        self.sym = symlink.filter_factory({})(app)
        self.sym.logger = self.swift.logger
        self.ov = ov.ObjectVersioningMiddleware(self.sym, {})
        self.ov.logger = self.swift.logger
        self.slo = slo.filter_factory({'rate_limit_under_size': '0'})(self.ov)
        self.slo.logger = self.swift.logger
        return super(TestVersionedMpuGETorHEAD, self)._wrap_app(self.slo)

    def setUp(self):
        # this will call our _wrap_app
        super(TestVersionedMpuGETorHEAD, self).setUp()
        self.ts = make_timestamp_iter()
        self.swift.register('HEAD', '/v1/AUTH_test/bucket+segments',
                            swob.HTTPNoContent, {}, None)
        versions_container = get_reserved_name('versions', 'bucket')
        self.swift.register(
            'HEAD', '/v1/AUTH_test/bucket', swob.HTTPNoContent, {
                ov.SYSMETA_VERSIONS_CONT: versions_container,
                ov.SYSMETA_VERSIONS_ENABLED: True,
            }, None)
        self.swift.register('HEAD', '/v1/AUTH_test/%s' % versions_container,
                            swob.HTTPNoContent, {}, None)
        num_versions = 3
        self.version_ids = []
        for v in range(num_versions):
            upload_id = 'X%s' % v
            num_segments = 3 + v
            manifest_headers, manifest_json = _prepare_mpu(
                self.swift, self.ts, upload_id, num_segments)
            version_ts = next(self.ts)
            # add in a little user-meta to keep versions stright
            manifest_version_headers = dict(manifest_headers, **{
                'x-object-meta-user-notes': 'version%s' % v,
                'x-backend-timestamp': version_ts.internal,
            })
            self.version_ids.append(version_ts.normal)
            obj_version_path = get_reserved_name('mpu', (~version_ts).normal)
            self.swift.register(
                'GET', '/v1/AUTH_test/%s/%s' % (
                    versions_container, obj_version_path),
                swob.HTTPOk, manifest_version_headers,
                manifest_json.encode('ascii'))
        # TODO: make a current version symlink
        symlink_target = '%s/%s' % (versions_container, obj_version_path)
        slo_etag = manifest_headers['X-Object-Sysmeta-Slo-Etag']
        s3_etag = manifest_headers['X-Object-Sysmeta-S3Api-Etag']
        symlink_target_etag = json_md5 = manifest_headers['Etag']
        symlink_target_bytes = manifest_headers['X-Object-Sysmeta-Slo-Size']
        manifest_symlink_headers = dict(manifest_headers, **{
            'X-Object-Sysmeta-Container-Update-Override-Etag':
            '%s; s3_etag=%s; slo_etag=%s; symlink_target=%s; '
            'symlink_target_etag=%s; symlink_target_bytes=%s' % (
                json_md5, s3_etag, slo_etag, symlink_target,
                symlink_target_etag, symlink_target_bytes),
            'X-Object-Sysmeta-Allow-Reserved-Names': 'true',
            'X-Object-Sysmeta-Symlink-Target': symlink_target,
            'X-Object-Sysmeta-Symlink-Target-Bytes': str(symlink_target_bytes),
            'X-Object-Sysmeta-Symlink-Target-Etag': symlink_target_etag,
            'X-Object-Sysmeta-Symloop-Extend': 'true',
            'X-Object-Sysmeta-Versions-Symlink': 'true',
        })
        self.swift.register(
            'GET', '/v1/AUTH_test/bucket/mpu', swob.HTTPOk,
            manifest_symlink_headers, '')
        self.s3_acl = False

    def test_mpu_GET_version(self):
        req = swob.Request.blank('/bucket/mpu', params={
            'versionId': self.version_ids[0],
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status, '200 OK')
        self.assertEqual(headers['x-amz-meta-user-notes'], 'version0')
        self.assertEqual(headers['Content-Length'], '30')
        self.assertEqual(body, b'aaaaabbbbbbbbbbccccccccccccccc')
        expected_calls = [
            ('HEAD', '/v1/AUTH_test'),
            ('HEAD', '/v1/AUTH_test/bucket'),
            ('HEAD', '/v1/AUTH_test/\x00versions\x00bucket'),
            ('GET', '/v1/AUTH_test/\x00versions\x00bucket/\x00mpu\x00%s'
             '?version-id=%s' % (
                 (~utils.Timestamp(self.version_ids[0])).normal,
                 self.version_ids[0])),
            ('HEAD', '/v1/AUTH_test/bucket+segments'),
            ('GET', '/v1/AUTH_test/bucket+segments/mpu/X0/1'
             '?multipart-manifest=get'),
            ('GET', '/v1/AUTH_test/bucket+segments/mpu/X0/2'
             '?multipart-manifest=get'),
            ('GET', '/v1/AUTH_test/bucket+segments/mpu/X0/3'
             '?multipart-manifest=get')
        ]
        if self.s3_acl:
            expected_calls.insert(3, (
                'HEAD', '/v1/AUTH_test/\x00versions\x00bucket/\x00mpu\x00%s'
                '?version-id=%s' % (
                    (~utils.Timestamp(self.version_ids[0])).normal,
                    self.version_ids[0])
            ))
        self.assertEqual(self.swift.calls, expected_calls)

    def test_mpu_GET_last_version(self):
        req = swob.Request.blank('/bucket/mpu', headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status, '200 OK')
        self.assertEqual(headers['x-amz-meta-user-notes'], 'version2')
        self.assertEqual(headers['Content-Length'], '75')
        expected_calls = [
            ('HEAD', '/v1/AUTH_test'),
            ('HEAD', '/v1/AUTH_test/bucket'),
            ('HEAD', '/v1/AUTH_test/\x00versions\x00bucket'),
            ('GET', '/v1/AUTH_test/bucket/mpu'),
            ('GET', '/v1/AUTH_test/\x00versions\x00bucket/\x00mpu\x00%s' % (
                ~utils.Timestamp(self.version_ids[2])).normal),
            ('HEAD', '/v1/AUTH_test/bucket+segments'),
            ('GET', '/v1/AUTH_test/bucket+segments/mpu/X2/1'
             '?multipart-manifest=get'),
            ('GET', '/v1/AUTH_test/bucket+segments/mpu/X2/2'
             '?multipart-manifest=get'),
            ('GET', '/v1/AUTH_test/bucket+segments/mpu/X2/3'
             '?multipart-manifest=get'),
            ('GET', '/v1/AUTH_test/bucket+segments/mpu/X2/4'
             '?multipart-manifest=get'),
            ('GET', '/v1/AUTH_test/bucket+segments/mpu/X2/5'
             '?multipart-manifest=get'),
        ]
        if self.s3_acl:
            # the pre-flight head on version marker get's symlinked; but I
            # think maybe symlink makes metadata addative?
            expected_calls = expected_calls[:3] + [
                ('HEAD', '/v1/AUTH_test/bucket/mpu'),
                ('HEAD', '/v1/AUTH_test/\x00versions\x00bucket/\x00mpu\x00'
                 '%s' % (~utils.Timestamp(self.version_ids[2])).normal),
            ] + expected_calls[3:]
        self.assertEqual(expected_calls, self.swift.calls)

    def test_mpu_HEAD_last_version(self):
        req = swob.Request.blank('/bucket/mpu', method='HEAD', headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status, '200 OK')
        self.assertEqual(headers['x-amz-meta-user-notes'], 'version2')
        self.assertEqual(headers['Content-Length'], '75')
        self.assertEqual([
            ('HEAD', '/v1/AUTH_test'),
            ('HEAD', '/v1/AUTH_test/bucket'),
            ('HEAD', '/v1/AUTH_test/\x00versions\x00bucket'),
            ('HEAD', '/v1/AUTH_test/bucket/mpu'),
            ('HEAD', '/v1/AUTH_test/\x00versions\x00bucket/\x00mpu\x00%s' % (
                ~utils.Timestamp(self.version_ids[2])).normal),
        ], self.swift.calls)

    def test_mpu_HEAD_version(self):
        req = swob.Request.blank('/bucket/mpu', method='HEAD', params={
            'versionId': self.version_ids[1],
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status, '200 OK')
        self.assertEqual(headers['x-amz-meta-user-notes'], 'version1')
        self.assertEqual(headers['Content-Length'], '50')
        self.assertEqual(body, b'')
        self.assertEqual([
            ('HEAD', '/v1/AUTH_test'),
            ('HEAD', '/v1/AUTH_test/bucket'),
            ('HEAD', '/v1/AUTH_test/\x00versions\x00bucket'),
            ('HEAD', '/v1/AUTH_test/\x00versions\x00bucket/\x00mpu\x00%s'
             '?version-id=%s' % (
                 (~utils.Timestamp(self.version_ids[1])).normal,
                 self.version_ids[1])),
        ], self.swift.calls)

    def test_mpu_GET_version_part_num(self):
        req = swob.Request.blank('/bucket/mpu', params={
            'versionId': self.version_ids[2],
            'partNumber': 5,
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status, '206 Partial Content')
        self.assertEqual(headers['x-amz-meta-user-notes'], 'version2')
        self.assertEqual(headers['Content-Length'], '25')
        self.assertEqual(body, b'e' * 25)
        expected_calls = [
            ('HEAD', '/v1/AUTH_test'),
            ('HEAD', '/v1/AUTH_test/bucket'),
            ('HEAD', '/v1/AUTH_test/\x00versions\x00bucket'),
            ('GET', '/v1/AUTH_test/\x00versions\x00bucket/\x00mpu\x00%s'
             '?part-number=5&version-id=%s' % (
                 (~utils.Timestamp(self.version_ids[2])).normal,
                 self.version_ids[2])),
            ('HEAD', '/v1/AUTH_test/bucket+segments'),
            ('GET', '/v1/AUTH_test/bucket+segments/mpu/X2/5'
             '?multipart-manifest=get'),
        ]
        if self.s3_acl:
            expected_calls.insert(3, (
                'HEAD', '/v1/AUTH_test/\x00versions\x00bucket/\x00mpu\x00%s'
                '?version-id=%s' % (
                    (~utils.Timestamp(self.version_ids[2])).normal,
                    self.version_ids[2])
            ))
        self.assertEqual(expected_calls, self.swift.calls)

    def test_mpu_HEAD_version_part_num(self):
        req = swob.Request.blank('/bucket/mpu', method='HEAD', params={
            'versionId': self.version_ids[2],
            'partNumber': 3,
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status, '206 Partial Content')
        self.assertEqual(headers['x-amz-meta-user-notes'], 'version2')
        self.assertEqual(headers['Content-Length'], '15')
        self.assertEqual(body, b'')
        self.assertEqual(self.swift.calls, [
            ('HEAD', '/v1/AUTH_test'),
            ('HEAD', '/v1/AUTH_test/bucket'),
            ('HEAD', '/v1/AUTH_test/\x00versions\x00bucket'),
            ('HEAD', '/v1/AUTH_test/\x00versions\x00bucket/\x00mpu\x00%s'
             '?part-number=3&version-id=%s' % (
                 (~utils.Timestamp(self.version_ids[2])).normal,
                 self.version_ids[2])),
            ('GET', '/v1/AUTH_test/\x00versions\x00bucket/\x00mpu\x00%s'
             '?part-number=3&version-id=%s' % (
                 (~utils.Timestamp(self.version_ids[2])).normal,
                 self.version_ids[2])),
        ])

    def test_mpu_GET_last_version_part_num(self):
        req = swob.Request.blank('/bucket/mpu', params={
            'partNumber': 4,
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status, '206 Partial Content')
        self.assertEqual(headers['x-amz-meta-user-notes'], 'version2')
        self.assertEqual(headers['Content-Length'], '20')
        self.assertEqual(body, b'd' * 20)
        expected_calls = [
            ('HEAD', '/v1/AUTH_test'),
            ('HEAD', '/v1/AUTH_test/bucket'),
            ('HEAD', '/v1/AUTH_test/\x00versions\x00bucket'),
            ('GET', '/v1/AUTH_test/bucket/mpu?part-number=4'),
            ('GET', '/v1/AUTH_test/\x00versions\x00bucket/\x00mpu\x00%s'
             '?part-number=4' % (
                 ~utils.Timestamp(self.version_ids[2])).normal),
            ('HEAD', '/v1/AUTH_test/bucket+segments'),
            ('GET', '/v1/AUTH_test/bucket+segments/mpu/X2/4'
             '?multipart-manifest=get'),
        ]
        if self.s3_acl:
            expected_calls = expected_calls[:3] + [
                ('HEAD', '/v1/AUTH_test/bucket/mpu'),
                ('HEAD', '/v1/AUTH_test/\x00versions\x00bucket/\x00mpu\x00'
                 '%s' % (~utils.Timestamp(self.version_ids[2])).normal),
            ] + expected_calls[3:]
        self.assertEqual(expected_calls, self.swift.calls)

    def test_mpu_HEAD_last_version_part_num(self):
        req = swob.Request.blank('/bucket/mpu', method='HEAD', params={
            'partNumber': 5,
        }, headers={
            'Authorization': 'AWS test:tester:hmac',
            'Date': self.get_date_header()
        })
        status, headers, body = self.call_s3api(req)
        self.assertEqual(status, '206 Partial Content')
        self.assertEqual(headers['x-amz-meta-user-notes'], 'version2')
        self.assertEqual(headers['Content-Length'], '25')
        self.assertEqual(self.swift.calls, [
            ('HEAD', '/v1/AUTH_test'),
            ('HEAD', '/v1/AUTH_test/bucket'),
            ('HEAD', '/v1/AUTH_test/\x00versions\x00bucket'),
            ('HEAD', '/v1/AUTH_test/bucket/mpu?part-number=5'),
            ('HEAD', '/v1/AUTH_test/\x00versions\x00bucket/\x00mpu\x00%s'
             '?part-number=5' % (
                 ~utils.Timestamp(self.version_ids[2])).normal),
            ('GET', '/v1/AUTH_test/bucket/mpu?part-number=5'),
            ('GET', '/v1/AUTH_test/\x00versions\x00bucket/\x00mpu\x00%s'
             '?part-number=5' % (
                 ~utils.Timestamp(self.version_ids[2])).normal),
        ])


class TestVersionedMpuGETorHEADAcl(TestVersionedMpuGETorHEAD,
                                   S3ApiTestCaseAcl):

    def setUp(self):
        super(TestVersionedMpuGETorHEADAcl, self).setUp()
        object_headers = _gen_test_headers(
            self.default_owner, self.grants, 'object')
        for version_id in self.version_ids:
            # s3acl would add the default object ACL on PUT to each version
            version_path = '/v1/AUTH_test/\x00versions\x00bucket/' \
                '\x00mpu\x00%s' % (~utils.Timestamp(version_id)).normal
            self.swift.update_sticky_response_headers(
                version_path, object_headers)
        # this is used to flag insertion of expected HEAD pre-flight request of
        # object ACLs
        self.s3_acl = True
